import os

from openbiolink.graph_creation import graphCreationConfig as glob
from openbiolink.graph_creation.metadata_edge.edgeOntoMetadata import EdgeOntoMetadata
from openbiolink.graph_creation.metadata_infile import InMetaOntoUberonPartOf


class EdgeMetaAnatomyOntoPartOf(EdgeOntoMetadata):
    NAME = 'Onto - Anatomy_partOf_Anatomy'

    EDGE_INMETA_CLASS = InMetaOntoUberonPartOf

    def __init__(self, quality=None):
        super().__init__(is_directional=True,
                         edges_file_path=os.path.join(glob.IN_FILE_PATH, self.EDGE_INMETA_CLASS.CSV_NAME),
                         colindex1=self.EDGE_INMETA_CLASS.NODE1_COL, colindex2=self.EDGE_INMETA_CLASS.NODE2_COL,
                         edgeType=self.EDGE_INMETA_CLASS.EDGE_TYPE,
                         node1_type=self.EDGE_INMETA_CLASS.NODE1_TYPE, node2_type=self.EDGE_INMETA_CLASS.NODE2_TYPE)



# from openbiolink.graph_creationtypes.dbType import DbType
import os

# from openbiolink.graph_creationtypes.readerType import ReaderType
from openbiolink.graph_creation import graphCreationConfig as g
from openbiolink.graph_creation.file_reader.csvReader import CsvReader


class MyCsvReader(CsvReader):
    """ to create a new csv file reader:
          *) declare the corresponding DB_META_CLASS, readerType, as well as dbType
          *) for clearer structure, move class to corresponding module (and import in corresponding init)
          prior steps necessary:
          *) create DB_META_CLASS
          *) add readerType
          *) add dbType
    """

    DB_META_CLASS = None  # database metaclass here

    def __init__(self):
        super().__init__(
            in_path=os.path.join(g.O_FILE_PATH, self.DB_META_CLASS.OFILE_NAME),
            sep=None,  # custom separator here (optional, if necessary (e.g. '|'))
            cols=self.DB_META_CLASS.COLS,
            use_cols=self.DB_META_CLASS.FILTER_COLS,
            nr_lines_header=self.DB_META_CLASS.HEADER,
            dtypes=None,
            readerType=None,  # reader type here
            dbType=None  # database type here
        )



O_FILE_PATH = ""
IN_FILE_PATH = ""

O_FILE_FOLDER_NAME = 'o_files'
IN_FILE_FOLDER_NAME = 'in_files'
GRAPH_FILES_FOLDER_NAME = 'graph_files'

EDGES_FILE_PREFIX = "edges"
NODES_FILE_PREFIX = "nodes"

ID_NO_MAPPING_FILE_NAME = 'ids_no_mapping.tsv'
STATS_FILE_NAME = 'stats.txt'
TN_ID_NO_MAPPING_FILE_NAME = 'tn_ids_no_mapping.tsv'
TN_STATS_FILE_NAME = 'tn_stats.txt'



import logging
import sys
import urllib.error
import urllib.request


class FileDownloader():
    @staticmethod
    def download(url, o_file_path):
        opener = urllib.request.build_opener()
        opener.addheaders = [('User-agent', 'Mozilla/5.0')]
        urllib.request.install_opener(opener)
        try:
            urllib.request.urlretrieve(url, o_file_path)
        except urllib.error.HTTPError as err:
            logging.error('HTTP %s %s:  %s' % (err.code, err.msg, err.geturl()))
            sys.exit()



from openbiolink.edgeType import EdgeType


class Edge:
    def __init__(self, id1: str, type: EdgeType, id2: str, source: "", qScore=None):
        self.id1 = id1
        self.type = type
        self.id2 = id2
        self.source = source
        self.qScore = qScore

    def __eq__(self, other):
        if isinstance(other, Edge):
            return self.type == other.type and self.id1 == other.id1 and self.id2 == other.id2  # todo only if directional
        return False

    def __hash__(self):
        return hash((self.id1, self.type, self.id2))

    def __iter__(self):
        return iter([self.id1, self.type, self.id2, self.qScore])

    def to_sub_rel_obj_list(self):
        return iter([self.id1, self.type, self.id2])



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile import InMetaEdgeStitchExpression
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class EdgeStitchExpressionProcessor(FileProcessor):
    IN_META_CLASS = InMetaEdgeStitchExpression

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_EDGE_STITCH_ACTION,
                         infileType=InfileType.IN_EDGE_STITCH_EXPRESSION, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_preprocessing(self, data):
        # only drug -> protein connections of single compounds (no merged)
        drug_protein = data.item_id_a.str.startswith('CIDs')
        mode = data['mode'] == 'expression'
        data = data[drug_protein & mode]
        self.stitch_to_pubchem_id(data, self.use_cols.index('item_id_a'))
        return data



from enum import Enum


class RankMetricType(Enum):
    HITS_AT_K = 'hits@K'
    HITS_AT_K_UNFILTERED = 'unfiltered hits@k'
    MRR = 'MRR'
    MRR_UNFILTERED = 'unfiltered MRR'


class ThresholdMetricType(Enum):
    ROC = 'ROC'
    ROC_AUC = 'ROC AUC'
    PR_REC_CURVE = 'precision-recall curve (PRC)'
    PR_AUC = 'PRC AUC'



from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeBgeeExpr import InMetaEdgeBgeeExpr
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeBgeeNoExpr import InMetaEdgeBgeeNoExpr
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeBgeeOverExpr import InMetaEdgeBgeeOverExpr
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeBgeeUnderExpr import InMetaEdgeBgeeUnderExpr
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeCdtPath import InMetaEdgeCdtPath
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeDisGeNet import InMetaEdgeDisGeNet
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeDrugCentral import InMetaEdgeDrugCentral
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeDrugCentralContraInd import InMetaEdgeDrugCentralContraInd
# from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeHpa       import InMetaEdgeHpa
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeGo import InMetaEdgeGo
# from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeSiderInd  import InMetaEdgeSiderInd
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeHpoDis import InMetaEdgeHpoDis
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeHpoGene import InMetaEdgeHpoGene
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeSiderSe import InMetaEdgeSiderSe
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitch import InMetaEdgeStitch
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchActivation import InMetaEdgeStitchActivation
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchBindAct import InMetaEdgeStitchBindAct
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchBindInh import InMetaEdgeStitchBindInh
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchBinding import InMetaEdgeStitchBinding
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchCatalysis import InMetaEdgeStitchCatalysis
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchExpression import InMetaEdgeStitchExpression
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchInhibition import InMetaEdgeStitchInhibition
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchPredBind import InMetaEdgeStitchPredBind
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStitchReaction import InMetaEdgeStitchReaction
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeString import InMetaEdgeString
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringActivation import InMetaEdgeStringActivation
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringBindAct import InMetaEdgeStringBindAct
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringBindInh import InMetaEdgeStringBindInh
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringBinding import InMetaEdgeStringBinding
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringCatalysis import InMetaEdgeStringCatalysis
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringExpression import InMetaEdgeStringExpression
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringInhibition import InMetaEdgeStringInhibition
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringPtmod import InMetaEdgeStringPtmod
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeStringReaction import InMetaEdgeStringReaction
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeTnHpoDis import InMetaEdgeTnHpoDis



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile import InMetaOntoUberonPartOf
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class OntoUberonPartOfProcessor(FileProcessor):
    IN_META_CLASS = InMetaOntoUberonPartOf

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_ONTO_UBERON,
                         infileType=InfileType.IN_ONTO_UBERON_PART_OF, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_postprocessing(self, data):
        # bgee is only mapping on CL and UBERON terms
        data = data[data['ID'].str.startswith('UBERON:') | data['ID'].str.startswith('CL:')]
        data = data[data['PART_OF'].str.startswith('UBERON:') | data['PART_OF'].str.startswith('CL:')]
        return data



from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapDisGeNet import InMetaMapDisGeNet
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapDrugCentralPubchem import InMetaMapDrugCentralPubchem
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoDoAltid import InMetaMapOntoDoAltid
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoDoOmim import InMetaMapOntoDoOmim
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoDoUmls import InMetaMapOntoDoUmls
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoGoAltid import InMetaMapOntoGoAltid
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoHpoAltid import InMetaMapOntoHpoAltid
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoHpoUmls import InMetaMapOntoHpoUmls
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoUberonAltid import InMetaMapOntoUberonAltid
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapString import InMetaMapString
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapUniEnsNcbi import InMetaMapUniEnsNcbi
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapUniUniNcbi import InMetaMapUniUniNcbi



from openbiolink.graph_creation.metadata_db_file.dbMetadata import DbMetadata
from openbiolink.graph_creation.metadata_db_file.edge import *
from openbiolink.graph_creation.metadata_db_file.mapping import *
from openbiolink.graph_creation.metadata_db_file.onto import *



import os

from openbiolink.graph_creation import graphCreationConfig as g
# from openbiolink.graph_creationtypes.dbType import DbType
# from openbiolink.graph_creationtypes.readerType import ReaderType
from openbiolink.graph_creation.file_reader.oboReader import OboReader


class MyOntoReader(OboReader):
    """ to create a new obo file reader:
       *) declare the corresponding DB_META_CLASS, readerType, as well as dbType
       *) for clearer structure, move class to corresponding module (and import in corresponding init)
       prior steps necessary:
       *) create DB_META_CLASS
       *) add readerType
       *) add dbType
        """

    DB_META_CLASS = None  # database metaclass here

    def __init__(self):
        super().__init__(
            in_path=os.path.join(g.O_FILE_PATH, self.DB_META_CLASS.OFILE_NAME),
            quadruple_list=self.DB_META_CLASS.QUADRUPLES,
            readerType=None,  # reader type here
            dbType=None  # database type here
        )



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaEdgeStitchBinding(InfileMetadata):
    CSV_NAME = "DB_STITCH_drug_binding_gene.csv"
    USE_COLS = ['item_id_a', 'item_id_b', 'score']
    NODE1_COL = 0
    NODE2_COL = 1
    QSCORE_COL = 2
    NODE1_TYPE = NodeType.DRUG
    NODE2_TYPE = NodeType.GENE
    EDGE_TYPE = EdgeType.DRUG_BINDING_GENE
    INFILE_TYPE = InfileType.IN_EDGE_STITCH_BINDING

    MAPPING_SEP = None

    def __init__(self):
        super().__init__(csv_name=InMetaEdgeStitchBinding.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaEdgeStitchBinding.INFILE_TYPE)



from enum import Enum


class ReaderType(Enum):
    READER_ONTO_GO = 0
    READER_ONTO_DO = 1
    READER_ONTO_HPO = 2
    READER_ONTO_UBERON = 3

    READER_EDGE_CDT_PATH = 100
    READER_EDGE_DISGENET = 101
    READER_EDGE_GO = 102
    READER_EDGE_HPA = 103
    READER_EDGE_HPO_DIS = 104
    READER_EDGE_HPO_GENE = 105
    READER_EDGE_SIDER_IND = 106
    READER_EDGE_SIDER_SE = 107
    READER_EDGE_STITCH = 108
    READER_EDGE_STRING = 109
    READER_EDGE_DRUGCENTRAL_IND = 110
    READER_EDGE_BGEE = 111
    READER_EDGE_BGEE_DIFF = 112
    READER_EDGE_STITCH_ACTION = 113
    READER_EDGE_STRING_ACTION = 114
    READER_EDGE_TN_HPO_DIS = 115

    READER_MAP_DISGENET = 10
    READER_MAP_STRING = 11
    READER_MAP_UNIPROT = 12
    READER_MAP_DRUGCENTRAL_PUBCHEM = 13



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapOntoGoAltid(InfileMetadata):
    CSV_NAME = "DB_ONTO_mapping_GO_alt_id.csv"
    USE_COLS = ['ID', 'ALT_ID']
    SOURCE_COL = 1
    TARGET_COL = 0
    MAPPING_SEP = ';'  # ';' sep is created while parsing
    INFILE_TYPE = InfileType.IN_MAP_ONTO_GO_ALT_ID

    def __init__(self):
        super().__init__(csv_name=InMetaMapOntoGoAltid.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapOntoGoAltid.INFILE_TYPE)



import tkinter as tk
from tkinter import ttk
from tkinter.scrolledtext import ScrolledText


class ConfirmFrame(tk.Frame):
    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller

        title = tk.Label(self, text="Confirm and Start", font=controller.title_font)

        global_variable_panel = tk.LabelFrame(self, text='Global Config')
        self.global_message = ScrolledText(global_variable_panel, height=2)

        gc_variable_panel = tk.LabelFrame(self, text='Graph Creation Options')
        self.gc_message = ScrolledText(gc_variable_panel, height=3)

        tts_variable_panel = tk.LabelFrame(self, text='Train Test Split Options')
        self.tts_message = ScrolledText(tts_variable_panel, height=3)

        eval_variable_panel = tk.LabelFrame(self, text='Training and Evaluation Options')
        self.eval_message = ScrolledText(eval_variable_panel, height=3)

        buttons_panel = tk.Frame(self)
        next_button = tk.Button(buttons_panel, text="Start", command=lambda: self.controller.start(), height=1,
                                width=15)
        prev_button = tk.Button(buttons_panel, text="Back", command=lambda: self.controller.show_previous_frame(),
                                height=1, width=15)

        title.pack(side="top", fill="x", pady=10)
        global_variable_panel.pack(side='top', fill='both', expand=True, padx=15)
        self.global_message.pack(side='left', fill='both', expand=True, anchor='w')
        gc_variable_panel.pack(side='top', fill='both', expand=True, padx=15)
        self.gc_message.pack(side='left', fill='both', expand=True, anchor='w')
        tts_variable_panel.pack(side='top', fill='both', expand=True, padx=15)
        self.tts_message.pack(side='left', fill='both', expand=True, anchor='w')
        eval_variable_panel.pack(side='top', fill='both', expand=True, padx=15)
        self.eval_message.pack(side='left', fill='both', expand=True, anchor='w')

        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 0), padx=10, anchor='s')
        buttons_panel.pack(side='bottom', padx=15, fill='x')
        prev_button.pack(side='left', anchor='w', pady=(5, 10))
        next_button.pack(side='right', anchor='e', pady=(5, 10))

    def update(self):
        self.global_message.delete(1.0, 'end')  # fixme why is controller not updated?
        self.gc_message.delete(1.0, 'end')
        self.tts_message.delete(1.0, 'end')
        self.eval_message.delete(1.0, 'end')
        self.gc_message.insert('end', self.args_list_to_string(self.controller.ARGS_LIST_GRAPH_CREATION))
        self.global_message.insert('end', self.args_list_to_string(self.controller.ARGS_LIST_GLOBAL))
        self.tts_message.insert('end', self.args_list_to_string(self.controller.ARGS_LIST_TRAIN_TEST_SPLIT))
        self.eval_message.insert('end', self.args_list_to_string(self.controller.ARGS_LIST_EVAL))

    def args_list_to_string(self, arg_list):
        params_gc_string = ''
        for param in arg_list:
            if param.startswith('--'):
                params_gc_string = params_gc_string + '\n\t'
            params_gc_string = params_gc_string + ' ' + param
        return params_gc_string



from openbiolink.graph_creation.metadata_db_file.mapping.dbMetadataMapping import DbMetadataMapping
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaMapUniprot(DbMetadataMapping):
    NAME = 'Mapping - Uniprot - Gene (ensembl/uniprot -> ncbi entrez)'
    URL = "ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/by_organism/HUMAN_9606_idmapping_selected.tab.gz"
    OFILE_NAME = "Uniprot_mapping_gene.tab.gz"
    COLS = ['UniProtKB-AC', 'UniProtKB-ID', 'GeneID',
            'RefSeq', 'GI', 'PDB', 'GO', 'UniRef100',
            'UniRef90', 'UniRef50', 'UniParc', 'PIR',
            'NCBI-taxon', 'MIM', 'UniGene', 'PubMed',
            'EMBL', 'EMBL-CDS', 'Ensembl', 'Ensembl_TRS',
            'Ensembl_PRO', 'Additional PubMed']
    FILTER_COLS = ['Ensembl', 'GeneID', 'UniProtKB-AC']
    HEADER = 0
    DTYPES = {'Ensembl': str, 'GeneID': str, 'UniProtKB-AC': str}
    DB_TYPE = DbType.DB_MAP_UNIPROT

    def __init__(self):
        super().__init__(url=DbMetaMapUniprot.URL,
                         ofile_name=DbMetaMapUniprot.OFILE_NAME,
                         dbType=DbMetaMapUniprot.DB_TYPE)



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeCtdPath(DbMetadataEdge):
    NAME = 'Edge - CTD - Gene Pathway'

    URL = "http://ctdbase.org/reports/CTD_genes_pathways.tsv.gz"
    OFILE_NAME = "CDT_gene_pathway.tsv.gz"
    COLS = ['geneSymb', 'geneID', 'pathName', 'pathID']
    FILTER_COLS = ['geneID', 'pathID']
    HEADER = 29
    DB_TYPE = DbType.DB_EDGE_CDT_PATH

    def __init__(self):
        super().__init__(url=DbMetaEdgeCtdPath.URL,
                         ofile_name=DbMetaEdgeCtdPath.OFILE_NAME,
                         dbType=DbMetaEdgeCtdPath.DB_TYPE)



import json
import os
import pickle

import numpy as np
import pykeen.constants as keenConst
import pykeen.utilities.pipeline as pipeline
import torch
import torch.optim as optim
from tqdm import tqdm

import openbiolink.evaluation.evalConfig as evalConst
import openbiolink.train_test_set_creation.ttsConfig as ttsConst
from openbiolink import globalConfig as globConst
from openbiolink.evaluation.models.model import Model


## ** code adapted from pykeen: https://github.com/SmartDataAnalytics/PyKEEN **


class PyKeen_BasicModel(Model):
    def __init__(self, config=None):
        super().__init__()
        if type(config) == str:
            with open(config) as file:
                content = file.read()
                self.config = json.loads(content)
        else:
            self.config = config
        self.device = torch.device('cpu')

        self.config[keenConst.NUM_ENTITIES] = None
        self.config[keenConst.NUM_RELATIONS] = None

        self.kge_model = None

    @staticmethod
    def load_model(config):  # todo
        if type(config) == str:
            with open(config) as file:
                content = file.read()
                model_dir = os.path.abspath(os.path.join(config, os.pardir))
        else:
            model_dir = None
        with open(os.path.join(model_dir, 'configuration.json')) as f:
            config = json.load(f)

        kge_model = pipeline.get_kge_model(config=config)

        if model_dir is not None:
            path_to_model = os.path.join(model_dir, 'trained_model.pkl')
            kge_model.load_state_dict(torch.load(path_to_model))
        return kge_model

    def train(self, pos_triples: np.array, neg_triples: np.array):

        all_triples = np.concatenate((pos_triples, neg_triples))

        # testme
        self.config[keenConst.NUM_ENTITIES] = len(np.unique(np.concatenate((all_triples[:, 0], all_triples[:, 2]))))
        self.config[keenConst.NUM_RELATIONS] = len(np.unique(all_triples[:, 1]))

        ## prepare model
        self.kge_model = pipeline.get_kge_model(config=self.config)
        self.kge_model = self.kge_model.to(self.device)

        optimizer = optim.SGD(self.kge_model.parameters(), lr=self.config[keenConst.LEARNING_RATE])
        loss_per_epoch = []
        num_pos_examples = pos_triples.shape[0]
        num_neg_examples = neg_triples.shape[0]

        ### train model
        for _ in tqdm(range(self.config[keenConst.NUM_EPOCHS])):
            # create batches
            indices_pos = np.arange(num_pos_examples)
            np.random.shuffle(indices_pos)
            pos_triples = pos_triples[indices_pos]
            pos_batches = self._split_list_in_batches(input_list=pos_triples,
                                                      batch_size=self.config["batch_size"])
            indices_neg = np.arange(num_neg_examples)
            np.random.shuffle(indices_neg)
            neg_triples = neg_triples[indices_neg]
            neg_batches = self._split_list_in_batches(input_list=neg_triples,
                                                      batch_size=self.config["batch_size"])
            current_epoch_loss = 0.

            for pos_batch, neg_batch in tqdm(zip(pos_batches, neg_batches), total=len(neg_batches)):
                current_batch_size = len(pos_batch)

                # if not len(pos_batch) == len(neg_batch):
                #    raise RuntimeError('Pos and neg batches are not the same length')

                pos_batch_tensor = torch.tensor(pos_batch, dtype=torch.long, device=self.device)
                neg_batch_tensor = torch.tensor(neg_batch, dtype=torch.long, device=self.device)
                # Recall that torch *accumulates* gradients. Before passing in a
                # new instance, you need to zero out the gradients from the old instance
                optimizer.zero_grad()
                loss = self.kge_model(pos_batch_tensor, neg_batch_tensor)
                current_epoch_loss += (loss.item() * current_batch_size)

                loss.backward()
                optimizer.step()

            loss_per_epoch.append(current_epoch_loss / len(pos_triples))

        ### prepare results for output
        entity_to_embedding = {
            id: embedding.detach().cpu().numpy()
            for id, embedding in enumerate(self.kge_model.entity_embeddings.weight)
        }
        relation_to_embedding = {
            id: embedding.detach().cpu().numpy()
            for id, embedding in enumerate(self.kge_model.relation_embeddings.weight)
        }

        results = {
            "trained_model": self.kge_model,
            "loss_per_epoch": loss_per_epoch,
            "entity_to_embedding": entity_to_embedding,
            "relation_to_embedding": relation_to_embedding,
            "config": self.config
        }
        self.output_results(results)

        return self.kge_model

    def get_ranked_and_sorted_predictions(self, mapped_test_triples):
        # ** original code can be found in utilities/prediction_utils.py

        mapped_test_triples = torch.tensor(mapped_test_triples, dtype=torch.long, device=self.device)
        borders = []
        i = 0
        while True:  # todo #fixme explore further size limitations
            borders.append(min(i * 1000, len(mapped_test_triples)))
            if i * 1000 > len(mapped_test_triples):
                break
            i += 1
        predicted_scores = []
        [predicted_scores.extend(self.kge_model.predict(mapped_test_triples[borders[i - 1]:borders[i]])) for i, _ in
         enumerate(borders) if i > 0]  # todo maybe here a loop and batches?
        predicted_scores = np.array(predicted_scores)
        _, sorted_indices = torch.sort(torch.tensor(predicted_scores, dtype=torch.float),
                                       descending=False)

        sorted_indices = sorted_indices.cpu().numpy()
        ranked_test_triples = mapped_test_triples[sorted_indices, :]
        ranked_test_triples = np.array(ranked_test_triples)
        ranked_scores = np.reshape(predicted_scores[sorted_indices], newshape=(-1, 1))
        ranked_test_triples = np.column_stack((ranked_test_triples, ranked_scores))

        # ranked_triples = np.concatenate([ranked_subject_column,ranked_predicate_column, ranked_object_column , ranked_scores], axis=1)
        # ranked_triples = pandas.DataFrame({globConst.NODE1_ID_COL_NAME:[x for sublist in ranked_subject_column.tolist() for x in sublist],
        #                                   globConst.EDGE_TYPE_COL_NAME: [x for sublist in ranked_predicate_column.tolist() for x in sublist],
        #                                   globConst.NODE2_ID_COL_NAME:[x for sublist in ranked_object_column.tolist() for x in sublist],
        #                                   globConst.SCORE_COL_NAME : [x for sublist in ranked_scores.tolist() for x in sublist]})
        return ranked_test_triples, sorted_indices

    def output_results(self, results):
        output_directory = os.path.join(os.path.join(globConst.WORKING_DIR, evalConst.EVAL_OUTPUT_FOLDER_NAME),
                                        evalConst.MODEL_DIR)

        with open(os.path.join(output_directory, 'configuration.json'), 'w') as file:
            json.dump(results["config"], file, indent=2)

        with open(os.path.join(output_directory, 'entities_to_embeddings.pkl'), 'wb') as file:
            pickle.dump(results['entity_to_embedding'], file, protocol=pickle.HIGHEST_PROTOCOL)

        with open(os.path.join(output_directory, 'entities_to_embeddings.json'), 'w') as file:
            json.dump(
                {
                    key: list(map(float, array))
                    for key, array in results['entity_to_embedding'].items()
                },
                file,
                indent=2,
                sort_keys=True
            )

        if results[keenConst.RELATION_TO_EMBEDDING] is not None:
            with open(os.path.join(output_directory, 'relations_to_embeddings.pkl'), 'wb') as file:
                pickle.dump(results["relation_to_embedding"], file, protocol=pickle.HIGHEST_PROTOCOL)

            with open(os.path.join(output_directory, 'relations_to_embeddings.json'), 'w') as file:
                json.dump(
                    {
                        key: list(map(float, array))
                        for key, array in results["relation_to_embedding"].items()
                    },
                    file,
                    indent=2,
                    sort_keys=True,
                )

        with open(os.path.join(output_directory, 'losses.json'), 'w') as file:
            json.dump(results['loss_per_epoch'], file, indent=2, sort_keys=True)

    def output_model(self, path):
        torch.save(
            self.kge_model.state_dict(),
            path,
        )


class TransE_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):
        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="TransE",
                embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class TransR_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="TransR",
                embedding_dim=5,
                relation_embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class TransD_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="TransD",
                embedding_dim=5,
                relation_embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class TransH_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="TransH",
                embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                soft_constraints_weight=0.015625,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class DistMult_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="DistMult",
                embedding_dim=5,
                relation_embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class Unstructured_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="UM",
                embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class SE_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="SE",
                embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                normalization_of_entities=2,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)


class Rescal_PyKeen(PyKeen_BasicModel):

    def __init__(self, config=None):

        if not config:
            self.config = dict(
                training_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                               ttsConst.TRAIN_FILE_NAME),
                test_set_path=os.path.join(os.path.join(globConst.WORKING_DIR, ttsConst.TTS_FOLDER_NAME),
                                           ttsConst.TEST_FILE_NAME),
                execution_mode="Training_mode",
                random_seed=42,
                kg_embedding_model_name="RESCAL",
                embedding_dim=5,
                margin_loss=1,
                learning_rate=0.01,
                scoring_function=1,
                num_epochs=1,
                batch_size=256,
                preferred_device="cpu"
            )
        else:
            self.config = config
        super().__init__(self.config)



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeBgeeDiffExpr(DbMetadataEdge):
    NAME = 'Edge - Bgee - differential expression (over-/under-expressed)'
    URL = "ftp://ftp.bgee.org/bgee_v13/download/calls/diff_expr_calls/Homo_sapiens_diffexpr-anatomy-simple.tsv.zip"
    OFILE_NAME = "BGEE_overexpr.tsv.zip"
    COLS = ['gene_id', 'gene_name', 'anatomical_entity', 'anatomical_entity_name',
            'developmental_stage', 'developmental_stage_name', 'differential_expr', 'call_quality']
    FILTER_COLS = ['gene_id', 'anatomical_entity', 'differential_expr', 'call_quality']
    HEADER = 1
    DB_TYPE = DbType.DB_EDGE_BGEE_DIFF

    def __init__(self):
        super().__init__(url=DbMetaEdgeBgeeDiffExpr.URL,
                         ofile_name=DbMetaEdgeBgeeDiffExpr.OFILE_NAME,
                         dbType=DbMetaEdgeBgeeDiffExpr.DB_TYPE)



from openbiolink.graph_creation.metadata_edge.onto.edgeMetaAnatomyOntoIsA import EdgeMetaAnatomyOntoIsA
from openbiolink.graph_creation.metadata_edge.onto.edgeMetaAnatomyOntoPartOf import EdgeMetaAnatomyOntoPartOf
from openbiolink.graph_creation.metadata_edge.onto.edgeMetaDisOntoIsA import EdgeMetaDisOntoIsA
from openbiolink.graph_creation.metadata_edge.onto.edgeMetaGoOntoIsA import EdgeMetaGoOntoIsA
from openbiolink.graph_creation.metadata_edge.onto.edgeMetaGoOntoPartOf import EdgeMetaGoOntoPartOf
from openbiolink.graph_creation.metadata_edge.onto.edgeMetaPhenoOntoIsA import EdgeMetaPhenoOntoIsA



import os

from openbiolink.graph_creation import graphCreationConfig as glob
from openbiolink.graph_creation.metadata_edge.edgeRegularMetadata import EdgeRegularMetadata
from openbiolink.graph_creation.metadata_infile import InMetaEdgeStitchInhibition
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapString import InMetaMapString
from openbiolink.graph_creation.types.qualityType import QualityType


class EdgeMetaDrugInhibitionGene(EdgeRegularMetadata):
    NAME = 'Edge - Drug_inhibition_Gene'

    LQ_CUTOFF = 0
    MQ_CUTOFF = 400
    HQ_CUTOFF = 700

    EDGE_INMETA_CLASS = InMetaEdgeStitchInhibition
    MAP2_META_CLASS = InMetaMapString

    def __init__(self, quality: QualityType = None):
        edges_file_path = os.path.join(glob.IN_FILE_PATH, self.EDGE_INMETA_CLASS.CSV_NAME)
        mapping_file2 = os.path.join(glob.IN_FILE_PATH, self.MAP2_META_CLASS.CSV_NAME)
        super().__init__(is_directional=True,
                         edges_file_path=edges_file_path,
                         colindex1=self.EDGE_INMETA_CLASS.NODE1_COL, colindex2=self.EDGE_INMETA_CLASS.NODE2_COL,
                         edgeType=self.EDGE_INMETA_CLASS.EDGE_TYPE,
                         node1_type=self.EDGE_INMETA_CLASS.NODE1_TYPE, node2_type=self.EDGE_INMETA_CLASS.NODE2_TYPE,
                         colindex_qscore=self.EDGE_INMETA_CLASS.QSCORE_COL, quality=quality,
                         mapping2_file=mapping_file2, map2_sourceindex=self.MAP2_META_CLASS.SOURCE_COL,
                         map2_targetindex=self.MAP2_META_CLASS.TARGET_COL)



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeSiderSe import InMetaEdgeSiderSe
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class EdgeSiderSeProcessor(FileProcessor):
    IN_META_CLASS = InMetaEdgeSiderSe

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_EDGE_SIDER_SE,
                         infileType=InfileType.IN_EDGE_SIDER_SE, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_postprocessing(self, data):
        self.stitch_to_pubchem_id(data, self.use_cols.index('stitchID_stereo'))
        return data



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeString(DbMetadataEdge):
    NAME = 'Edge - STRING - Gene Gene'

    # current link can be extracted from https://stringdb-static.org/cgi/download.pl?sessionId=FuXJ9a0fkSMB&species_text=Homo+sapiens
    # URL = "https://stringdb-static.org/download/protein.links.v10.5/9606.protein.links.v10.5.txt.gz"
    URL = "https://stringdb-static.org/download/protein.links.v11.0/9606.protein.links.v11.0.txt.gz"
    OFILE_NAME = "STRING_gene_gene.txt.gz"
    COLS = ['string1', 'string2', 'qscore']
    FILTER_COLS = ['string1', 'string2', 'qscore']
    HEADER = 1
    DB_TYPE = DbType.DB_EDGE_STRING

    def __init__(self):
        super().__init__(url=DbMetaEdgeString.URL,
                         ofile_name=DbMetaEdgeString.OFILE_NAME,
                         dbType=DbMetaEdgeString.DB_TYPE)



from openbiolink.graph_creation.file_reader.edge.edgeBgeeDiffReader import EdgeBgeeDiffReader
from openbiolink.graph_creation.file_reader.edge.edgeBgeeReader import EdgeBgeeReader
from openbiolink.graph_creation.file_reader.edge.edgeCdtPathReader import EdgeCdtPathReader
from openbiolink.graph_creation.file_reader.edge.edgeDisGeNetReader import EdgeDisGeNetReader
from openbiolink.graph_creation.file_reader.edge.edgeDrugCentralReader import EdgeDrugCentralReader
from openbiolink.graph_creation.file_reader.edge.edgeGoReader import EdgeGoReader
# from openbiolink.graph_creation.file_reader.edge.edgeHpaReader import       EdgeHpaReader --> replaced by bgee
from openbiolink.graph_creation.file_reader.edge.edgeHpoDisReader import EdgeHpoDisReader
# from openbiolink.graph_creation.file_reader.edge.edgeSiderIndReader import  EdgeSiderIndReader --> replaced by drugcentral
from openbiolink.graph_creation.file_reader.edge.edgeHpoGeneReader import EdgeHpoGeneReader
from openbiolink.graph_creation.file_reader.edge.edgeSiderSeReader import EdgeSiderSeReader
from openbiolink.graph_creation.file_reader.edge.edgeStitchActionReader import EdgeStitchActionReader
from openbiolink.graph_creation.file_reader.edge.edgeStitchReader import EdgeStitchReader
from openbiolink.graph_creation.file_reader.edge.edgeStringActionReader import EdgeStringActionReader
from openbiolink.graph_creation.file_reader.edge.edgeStringReader import EdgeStringReader
from openbiolink.graph_creation.file_reader.edge.edgeTnHpoDisReader import EdgeTnHpoDisReader



import os

from openbiolink.graph_creation import graphCreationConfig as gcConst
from openbiolink.graph_creation.file_reader.csvReader import CsvReader
from openbiolink.graph_creation.metadata_db_file.mapping.dbMetaMapDisGeNet import DbMetaMapDisGeNet
from openbiolink.graph_creation.types.dbType import DbType
from openbiolink.graph_creation.types.readerType import ReaderType


class MapDisGeNetReader(CsvReader):
    DB_META_CLASS = DbMetaMapDisGeNet

    def __init__(self):
        super().__init__(
            in_path=os.path.join(gcConst.O_FILE_PATH, self.DB_META_CLASS.OFILE_NAME),
            sep='|',
            cols=self.DB_META_CLASS.COLS,
            use_cols=self.DB_META_CLASS.FILTER_COLS,
            nr_lines_header=self.DB_META_CLASS.HEADER,
            dtypes=None,
            readerType=ReaderType.READER_MAP_DISGENET,
            dbType=DbType.DB_MAP_DISGENET
        )



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapOntoDoUmls(InfileMetadata):
    CSV_NAME = "DB_ONTO_mapping_DO_UMLS.csv"
    USE_COLS = ['ID', 'UMLS']
    SOURCE_COL = 1
    TARGET_COL = 0
    MAPPING_SEP = ';'
    INFILE_TYPE = InfileType.IN_MAP_ONTO_DO_UMLS

    def __init__(self):
        super().__init__(csv_name=InMetaMapOntoDoUmls.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapOntoDoUmls.INFILE_TYPE)



from openbiolink.edgeType import EdgeType
from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.nodeType import NodeType


class InMetaEdgeBgeeUnderExpr(InfileMetadata):
    CSV_NAME = "DB_Bgee_gene_anatomy_underexpr.csv"
    USE_COLS = ['gene_id', 'anatomical_entity', 'call_quality']
    NODE1_COL = 0
    NODE2_COL = 1
    QSCORE_COL = 2
    NODE1_TYPE = NodeType.GENE
    NODE2_TYPE = NodeType.ANATOMY
    EDGE_TYPE = EdgeType.GENE_UNDEREXPRESSED_ANATOMY
    INFILE_TYPE = InfileType.IN_EDGE_BGEE_UNDEREXPR
    MAPPING_SEP = None

    def __init__(self):
        super().__init__(csv_name=InMetaEdgeBgeeUnderExpr.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaEdgeBgeeUnderExpr.INFILE_TYPE)



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapString(InfileMetadata):
    CSV_NAME = "DB_String_mapping_gene_ncbi_string.csv"
    USE_COLS = ['ncbiID', 'stringID']
    SOURCE_COL = 1
    TARGET_COL = 0
    MAPPING_SEP = '|'
    INFILE_TYPE = InfileType.IN_MAP_STRING

    def __init__(self):
        super().__init__(csv_name=InMetaMapString.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapString.INFILE_TYPE)



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeHpoGene(DbMetadataEdge):
    NAME = 'Edge - HPO - Gene Phenotype'
    URL = "http://compbio.charite.de/jenkins/job/hpo.annotations.monthly/lastSuccessfulBuild/artifact/annotation/ALL_SOURCES_ALL_FREQUENCIES_genes_to_phenotype.txt"
    OFILE_NAME = "HPO_gene_phenotype.tsv"
    COLS = ['geneID', 'geneSymb', 'hpoName', 'hpoID']
    FILTER_COLS = ['geneID', 'hpoID']
    HEADER = 1
    DB_TYPE = DbType.DB_EDGE_HPO_GENE

    def __init__(self):
        super().__init__(url=DbMetaEdgeHpoGene.URL,
                         ofile_name=DbMetaEdgeHpoGene.OFILE_NAME,
                         dbType=DbMetaEdgeHpoGene.DB_TYPE)



import csv
import inspect
import logging
import os
import urllib.request
from functools import reduce

import numpy as np
import pandas

from openbiolink import globalConfig as globConst
from openbiolink.edgeType import EdgeType


def get_leaf_subclasses(cls, class_set=None):
    """"
    Returns a set of all leaf subclasses (i.e. child classes that do not have child-classes on their own) of a specific class.
    If no subclasses exist, the class itself is returned.
    If cls=None, None is returned.

    Parameters
    ----------
       cls : class
           class of which leaf subclasses should be obtained
       class_set : set, optional
           this parameter is only meant for recursive calls
    Returns
    ----------
        class_set : set
            set of all leaf subclasses
    """
    if inspect.isclass(cls):
        if class_set is None:
            class_set = set()
        if len(cls.__subclasses__()) == 0:
            class_set.add(cls)
        else:
            class_set.union(x for c in cls.__subclasses__() for x in get_leaf_subclasses(c, class_set))
        return class_set
    return None


# def get_all_subclasses(self, cls):
#    return set(cls.__subclasses__()).union([x for c in cls.__subclasses__() for x in self.get_all_subclasses(c)])


def make_undir(data: pandas.DataFrame):
    """"
    Takes a DataFrame which contains directional edges with possible bidirectional edges and removes
    the bidirectional edges, so (A->B and B->A) becomes(A->B); further resets the index

    Parameters
    ----------
        data : pandas.DataFrame
            DataFrame with columns [id1, id2] or [id1, id2, score], containing directional edges
    Returns
    ----------
        new_data  : pandas.DataFrame
            DataFrame with columns [id1, id2] or [id1, id2, score] (depending on input), containing directional edges

    Examples
    ----------
    >>> d = pandas.DataFrame({'id1': list('abc'), 'id2': list('xcb')})
    >>> d
      id1 id2
    0   a   x
    1   b   c
    2   c   b
    >>> make_undir(d)
      id1 id2
    0   a   x
    1   b   c
    """

    temp_dic = {}
    cols = list(data)
    for row in data.itertuples():
        if len(cols) == 3:
            other_row = row[2] + row[1] + str(row[3])
        elif len(cols) == 2:
            other_row = row[2] + row[1]
        else:
            logging.warning(('removing bidirectional edges requires 2 or 3 (incl score) columns but cols are '.join(
                cols)) + 'edges are not removed')
            return data
        if len(cols) == 3:
            if other_row not in temp_dic.keys():
                this_row = row[1] + row[2] + str(row[3])
                temp_dic[this_row] = [row[1], row[2], row[3]]
        elif len(cols) == 2:
            if other_row not in temp_dic.keys():
                this_row = row[1] + row[2]
                temp_dic[this_row] = [row[1], row[2]]

    new_data = pandas.DataFrame.from_dict(temp_dic, columns=list(data), orient='index').reset_index(drop=True)
    return new_data


# source: https://gist.github.com/wonderbeyond/d293e7a2af1de4873f2d757edd580288#file-rgetattr-py
def rgetattr(obj, attr, *args):
    """"
        Parameters
        ----------
        Returns
        ----------
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split('.'))


def db_mapping_file_to_dic(mapping_file, map_source_col, map_target_col, sep=';'):
    """"
    creates a dic out of a metadata_db_file mapping file {source_id : [target_ids]}
    if multiple rows with the same source id exist, the target-id-list holds multiple elements
    if one row holds multiple targets (multiple targets in one target cell), the resulting string is treated as one element

    Parameters
    ----------
        mapping_file : str
            absolute path to the mapping file; mapping file must be a csv-like with one source column and one target column
        map_source_col : int
            column index of source column
        map_target_col : int
            column index of target column
        sep : str, optional, default = ';'
            separator of the mapping csv file

    Returns
    ----------
        mapping : dict

    Examples
    -----------
    if the mapping file looks like this

    | source | target |
    - - - - - - - - - -
    | s1     | t1     |
    | s2     | t2     |
    | s1     | t3     |
    | s3     | t4, t5 |

    the resulting dictionary looks like this
    {
    's1' : ['t1', 't3']
    's2' : ['t2']
    's3' : ['t4, t5']
    }
    """

    mapping = None
    if (mapping_file is not None):
        mapping = {}
        with open(mapping_file, mode="r") as mapping_content1:
            reader = csv.reader(mapping_content1, delimiter=sep)

            for row in reader:
                if row[map_target_col]:
                    if row[map_source_col] in mapping:
                        mapping[row[map_source_col]].append(row[map_target_col])
                    else:
                        mapping[row[map_source_col]] = [row[map_target_col]]
            mapping_content1.close()
    return mapping


def cls_list_to_dic(clsList, keyAttr, condition=None):
    """"
        creates an attribute dic out of a class list {keyAttribute_value : [classes]}

        Parameters
        ----------
            clsList : list
                list of classes that should be grouped by certain attribute; classes must all contain the ´keyAttr´,
                e.g. be all subclasses of class X, where X.keyAttr exists
            keyAttr : str
                attribute, by which the classes shall be grouped by
            condition : lambda
                condition for class that must hold in order to be considered in the grouping

        Returns
        ----------
            dic : dict
                dictionary, where the value of the keyAttr is the key, and a list of classes,
                where the key attribute has this value, is the value, i.e. {keyAttribute_value : [classes]}
    """

    if condition is None:
        condition = lambda a: True
    dic = {}
    for cls in clsList:
        if condition(cls):
            key = rgetattr(cls, keyAttr)
            if key in dic:
                dic[key].append(cls)
            elif key is not None:
                dic[key] = [cls]
    return dic


def url_exists(url):
    """"
    checks, whether an URL exists

        Parameters
        ----------
            url : str
                URL which should be checked
        Returns
        ----------
            bool
                True -> URL exists, False -> URL does not exist
    """
    opener = urllib.request.build_opener()
    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
    urllib.request.install_opener(opener)
    try:
        urllib.request.urlopen(url)
        return True
    except:
        return False


############# FROM TRAIN TEST SPLIT #################################

def get_diff(df1, df2, ignore_qscore=False, path=None):
    """"
    takes DataFrames with same column names and returns their differences
        Parameters
        ----------
            df1 : pandas.DataFrame
                left DataFrame of the merge
            df2 : pandas.DataFrame
                right DataFrame of the join
            ignore_qscore : bool
                whether the quality score columns should be ignored when merging the data frames
            path : str, optional
                if provided, the differences are saved as csv files in the provided path

        Returns
        ----------
            left_only : pandas.DataFrame
                DataFrame containing rows that only appear in df1, but not df2
            right_only : pandas.DataFrame
                DataFrame containing rows that only appear in df2, but not df1
    """
    df1['id1'] = df1['id1'].astype(str)
    df1['id2'] = df1['id2'].astype(str)
    df1['edgeType'] = df1['edgeType'].astype(str)
    df2['id1'] = df2['id1'].astype(str)
    df2['id2'] = df2['id2'].astype(str)
    df2['edgeType'] = df2['edgeType'].astype(str)
    if ignore_qscore:
        all_cols = globConst.COL_NAMES_SAMPLES
        cols = globConst.COL_NAMES_TRIPLES + [globConst.VALUE_COL_NAME]
        diff = pandas.merge(df1, df2,
                            how='outer',
                            left_on=cols,
                            right_on=cols,
                            indicator=True).loc[lambda x: x['_merge'] != 'both']
        left_only = diff.loc[lambda x: x._merge == 'left_only']
        left_only[globConst.QSCORE_COL_NAME] = left_only[globConst.QSCORE_COL_NAME + '_x']
        left_only = left_only[all_cols]
        left_only.drop_duplicates(inplace=True, keep=False)
        right_only = diff.loc[lambda x: x._merge == 'right_only']
        right_only[globConst.QSCORE_COL_NAME] = right_only[globConst.QSCORE_COL_NAME + '_y']
        right_only = right_only[all_cols]
        right_only.drop_duplicates(inplace=True, keep=False)

    else:
        diff = pandas.merge(df1, df2, how='outer', indicator=True).loc[lambda x: x['_merge'] != 'both']
        left_only = diff.loc[lambda x: x._merge == 'left_only'].drop(['_merge'], axis=1)
        right_only = diff.loc[lambda x: x._merge == 'right_only'].drop(['_merge'], axis=1)
    if path:
        left_only.to_csv(os.path.join(path, 'diff_left_only.csv'), sep='\t', index=False, header=False)
        right_only.to_csv(os.path.join(path, 'diff_right_only.csv'), sep='\t', index=False, header=False)
    return left_only, right_only


def remove_inconsistent_edges(df: pandas.DataFrame):
    """ removes edges that contain inconsistent information i.e. when an edge is present both as positive and negative example"""
    return df.drop_duplicates(
        subset=[globConst.NODE1_ID_COL_NAME, globConst.NODE2_ID_COL_NAME, globConst.EDGE_TYPE_COL_NAME],
        keep=False)


def remove_parent_duplicates_and_reverses(remain_set, remove_set):
    if not remain_set.empty and not remove_set.empty:
        remove_set_copy = remove_set.copy()
        remove_set_copy[globConst.EDGE_TYPE_COL_NAME] = remove_set_copy[globConst.EDGE_TYPE_COL_NAME] \
            .apply(lambda x: EdgeType[x].get_parent() if type(x) == str else x.get_parent())
        remove_set_copy.drop_duplicates(inplace=True, subset=[globConst.NODE1_ID_COL_NAME,
                                                              globConst.NODE2_ID_COL_NAME,
                                                              globConst.EDGE_TYPE_COL_NAME])
        remain_set, _ = get_diff(remain_set, remove_set_copy, ignore_qscore=True)
        remain_set.drop_duplicates(inplace=True, subset=[globConst.NODE1_ID_COL_NAME,
                                                         globConst.NODE2_ID_COL_NAME,
                                                         globConst.EDGE_TYPE_COL_NAME])
        remain_set = remove_reverse_edges(remain_set, remove_set_copy)
    return remain_set
    # testme


def remove_reverse_edges(remain_set, remove_set):
    """"
    removes the reverse edges of the remove_set from the remain_set (return = remain-remove^-1), sets may not contain duplicates!
        Parameters
        ----------
            remain_set : pandas.DataFrame
                DataFrame from which all reverse edges from remove_set shall be removed, columns #todo
            remove_set : pandas.DataFrame
                DataFrame holds all edges, whose reverse edges shall be removed in the remain_set, columns = #todo
        Returns
        ----------
            remain : pandas.DataFrame
                DataFame that contains no reverse edges from remove_set

    """
    # todo independent from globals
    if not remain_set.empty and not remove_set.empty:
        remain_set.reset_index(drop=True, inplace=True)
        remove_set.reset_index(drop=True, inplace=True)
        remove_set_copy = remove_set.copy()
        # changing columns of node1_id and node2_id
        cols_other_dir = list(remove_set_copy.columns)
        index_id1 = cols_other_dir.index(globConst.NODE1_ID_COL_NAME)
        index_id2 = cols_other_dir.index(globConst.NODE2_ID_COL_NAME)
        cols_other_dir[index_id1] = globConst.NODE2_ID_COL_NAME
        cols_other_dir[index_id2] = globConst.NODE1_ID_COL_NAME
        remove_set_copy.columns = cols_other_dir
        temp = pandas.merge(remain_set, remove_set_copy, how='left',
                            left_on=[globConst.NODE1_ID_COL_NAME, globConst.NODE2_ID_COL_NAME,
                                     globConst.EDGE_TYPE_COL_NAME],
                            right_on=[globConst.NODE1_ID_COL_NAME, globConst.NODE2_ID_COL_NAME,
                                      globConst.EDGE_TYPE_COL_NAME]
                            )
        temp.set_index(remain_set.index)
        temp.drop_duplicates(inplace=True)

        # todo value should always be different? are there examples with diff qscore?
        remove = temp[
            (temp['value_x'] == temp['value_y'])]  # if no match would be found in remove_set, value_y would be NaN
        remain = remain_set.drop(remove.index.values)
        return remain
    else:
        return remain_set


# nicetohave (2) more lvl transitivity
# def check_for_transitive_edges(df):
#    """"
#        Parameters
#        ----------
#        Returns
#        ----------
#    """
#    direct_child_dict = {}
#    for row in df[[globConst.NODE1_ID_COL_NAME, globConst.NODE2_ID_COL_NAME]].itertuples():
#        _, child, parent = row
#        if not parent in direct_child_dict.keys():
#            direct_child_dict[parent] = {child}
#        else:
#            direct_child_dict[parent].add(child)
#    # fix me check for cycles
#    all_children = set()
#    return (
#    [((child_set & direct_child_dict[c]), parent, c) for (parent, child_set) in direct_child_dict.items() for c in
#     child_set if c in direct_child_dict.keys() if (child_set & direct_child_dict[c])])
# def get_all_childs(self,direct_child_dict, current_parent, past_parents, all_children):
#    for child in direct_child_dict[current_parent]:
#        pass


def calc_corrupted_triples(pos_example, nodes, nodes_dic, filtered=False, path=None, pos_examples=None):
    """"
    calculates the corrupted triples (both corrupted heads as well as corrupted tails) for given true positive example
        Parameters
        ----------
            pos_examples : pandas.DataFrame #todo
                DataFrame with true examples of edges, columns = [node1_id, edgeType, node2_id]
            nodes : pandas.DataFrame #todo #todo
                DataFrame with all nodes of the graph columns = [node_id, node_type]
            filtered : bool, optional, default = True
                if filtered is true, filtered corrupted triples are created; that means, that only those corrupted triples,
                that do not exist in the graph, are kept
                if filtered is false, unfiltered corrupted triples are created; this means, that all corrupted triples are
                kept, even if they exist in the graph
            path : str, optional
                if provided, the corrupted triples are saved as csv files in the provided path; one file contains the corrupted heads, grouped by


        Returns
        ----------
    """
    filtered_corrupted_heads = None
    filtered_corrupted_tails = None
    head, relation, tail = pos_example
    head_node_type = nodes[np.where(nodes[:, 0] == head)][0, 1]
    corrupted_head_nodes = nodes_dic[head_node_type]
    corrupted_head_nodes = corrupted_head_nodes[corrupted_head_nodes != head]
    tail_node_type = nodes[np.where(nodes[:, 0] == tail)][0, 1]
    corrupted_tail_nodes = nodes_dic[tail_node_type]
    corrupted_tail_nodes = corrupted_tail_nodes[corrupted_tail_nodes != tail]

    # corrupting heads
    tail_array = np.full(len(corrupted_head_nodes), tail)
    relation_array = np.full(len(corrupted_head_nodes), relation)
    value_array = np.zeros(len(corrupted_head_nodes))
    unfiltered_corrupted_heads = np.column_stack((corrupted_head_nodes, relation_array, tail_array, value_array))

    # corrupting tails
    head_array = np.full(len(corrupted_tail_nodes), head)
    relation_array = np.full(len(corrupted_tail_nodes), relation)
    value_array = np.zeros(len(corrupted_tail_nodes))
    unfiltered_corrupted_tails = np.column_stack((head_array, relation_array, corrupted_tail_nodes, value_array))

    if filtered:
        filtered_corrupted_heads = _get_corrupted_examples(unfiltered_corrupted_heads[:, 0:3], pos_examples, True)
        filtered_corrupted_tails = _get_corrupted_examples(unfiltered_corrupted_tails[:, 0:3], pos_examples, True)

    # if path: #todo change to numpy #fixme what to do?
    #    all_corrupted_head = _group_corrupted_examples(unfiltered_corrupted_head_dict,
    #                                                   [evalConst.CORRUPTED_GROUP_COL_NAME]+list(pos_example)+[globConst.VALUE_COL_NAME])
    #    all_corrupted_tail = _group_corrupted_examples(unfiltered_corrupted_tail_dict,
    #                                                   [evalConst.CORRUPTED_GROUP_COL_NAME]+list(pos_example)+[globConst.VALUE_COL_NAME])
    #    all_corrupted_head.to_csv(os.path.join(path, evalConst.CORRUPTED_HEADS_FILE_NAME), sep='\t', index=False, header=False)
    #    all_corrupted_tail.to_csv(os.path.join(path, evalConst.CORRUPTED_TAILS_FILE_NAME), sep='\t', index=False, header=False)

    return unfiltered_corrupted_heads, unfiltered_corrupted_tails, filtered_corrupted_heads, filtered_corrupted_tails


def _get_corrupted_examples(corrupted_triples, pos_examples, filtered):
    """"
    Creates a dataset of corrupted triples according to the 'filtered' setting
        Parameters
        ----------
            corrupted_triples_df : pandas.DataFrame
                DataFrame with corrupted triples of one positive example, columns = [node1_id, edgeType, node2_id]
            pos_examples_df : pandas.DataFrame #todo
                DataFrame with all edges that are present in the graph (= positive examples), columns = [node1_id, edgeType, node2_id]
            filtered : bool
                if true, filtered setting is applied; that means, that only those corrupted triples, that do not exist in the graph, are kept
                if false, unfiltered setting is applied; this means, that all corrupted triples are kept, even if they exist in the graph

        Returns
        ----------
            corrupted_triples : pandas.DataFrame
                DataFrame contains all valid corrupted triples, incl value (1 for positive example, 0 for negative example)
                columns = [node1_id, edgeType, node2_id, value]
    """
    corrupted_triples_df = pandas.DataFrame(corrupted_triples, columns=globConst.COL_NAMES_TRIPLES)
    pos_examples_df = pandas.DataFrame(pos_examples, columns=globConst.COL_NAMES_TRIPLES)
    if filtered:
        corrupted_triples_df.reset_index(drop=True, inplace=True)
        true_neg_triples, _ = get_diff(corrupted_triples_df, pos_examples_df)
        corrupted_triples_df = true_neg_triples
        corrupted_triples_df[globConst.VALUE_COL_NAME] = 0
    else:
        corrupted_triples_df[globConst.VALUE_COL_NAME] = 0
        # corrupted_triples[globConst.VALUE_COL_NAME][neg_indices] = 0
    return corrupted_triples_df.values


def _group_corrupted_examples(corrupted_dict, col_names):
    """"
     Parameters
     ----------
        corrupted_dict : dictionary
            dictionary of all corrupted triples, triple (head_id, relation_id, tail_id) as key,
            pandas.DataFrame with columns [head, relation, tail, value]
        col_names : list
            list of column names for [group, head_id, relation_id, tail_id, value]
            where value determines whether the triple exists (1) or not (0) in the graph
     Returns
     ----------
        all_corrupted : pandas.DataFrame
            pandas DataFrame of all triples grouped by true positive triple
            columns are [group, head_id, relation_id, tail_id, value, tp]
            the true positive example (tp) is always the first on a group of corrupted triples
    """
    i = 0
    # col_names = col_names + ['tp']
    all_corrupted = pandas.DataFrame(columns=col_names)
    for key, df in sorted(corrupted_dict.items()):
        h, r, t = key
        true_example = pandas.DataFrame([[h, r, t]], columns=list(df))
        # true_example['tp'] = 1
        true_example[col_names[4]] = 1
        true_example[col_names[0]] = i
        # todo should there be an extra column for tp?
        # df['tp'] = 0
        df[col_names[0]] = i
        all_corrupted = all_corrupted.append(true_example)
        all_corrupted = all_corrupted.append(df)
        i += 1
    return all_corrupted


def create_mappings(elements):
    element_label_to_id = {element_label: id for id, element_label in enumerate(elements)}
    return element_label_to_id


def map_elements(elements, mapping):
    return np.vectorize(mapping.get)(elements)



import logging
import sys

import pandas as pd


class PostgresDumpParser():

    @staticmethod
    def table_to_df(f, table_name, cols=None):

        data_started = False
        df = None
        for line in f:
            line = line.strip()
            if data_started:
                if line.startswith('\.'):
                    break
                row = line.split('\t')
                row = [x.strip() for x in row]
                if df is None:
                    df = pd.DataFrame([row], columns=cols)
                else:
                    if len(row) == len(cols):
                        df = df.append(pd.DataFrame([row], columns=cols), ignore_index=True)
                    else:
                        logging.info('postgresParser -> row ignored: ', row)
            if line.startswith("COPY " + table_name):
                data_started = True
                entry_list = line.split('(')[1].split(')')[0].split(',')
                if cols is None:
                    cols = [x.strip() for x in entry_list]
                else:
                    if not (len(cols) == len(entry_list)):
                        logging.error(
                            "PostgresDumpParser: Number of columns provided for table %s are different from number found in table. %d provided, %d found" % (
                            table_name, len(cols), len(entry_list)))
                        sys.exit()
        return df



import os

from openbiolink.graph_creation import graphCreationConfig as glob
from openbiolink.graph_creation.metadata_edge.edgeRegularMetadata import EdgeRegularMetadata
from openbiolink.graph_creation.metadata_infile import InMetaMapOntoHpoAltid
from openbiolink.graph_creation.metadata_infile.edge.inMetaEdgeSiderSe import InMetaEdgeSiderSe
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapOntoHpoUmls import InMetaMapOntoHpoUmls
from openbiolink.graph_creation.types.qualityType import QualityType


class EdgeMetaDrugPheno(EdgeRegularMetadata):
    NAME = 'Edge - Drug_sideEffect_Phenotype'

    EDGE_INMETA_CLASS = InMetaEdgeSiderSe
    MAP2_META_CLASS = InMetaMapOntoHpoUmls
    MAP2_ALT_ID_META_CLASS = InMetaMapOntoHpoAltid

    def __init__(self, quality: QualityType = None):
        edges_file_path = os.path.join(glob.IN_FILE_PATH, self.EDGE_INMETA_CLASS.CSV_NAME)
        mapping_file2 = os.path.join(glob.IN_FILE_PATH, self.MAP2_META_CLASS.CSV_NAME)
        altid_mapping_file2 = os.path.join(glob.IN_FILE_PATH, self.MAP2_ALT_ID_META_CLASS.CSV_NAME)

        super().__init__(is_directional=True,
                         edges_file_path=edges_file_path,
                         colindex1=self.EDGE_INMETA_CLASS.NODE1_COL, colindex2=self.EDGE_INMETA_CLASS.NODE2_COL,
                         edgeType=self.EDGE_INMETA_CLASS.EDGE_TYPE,
                         node1_type=self.EDGE_INMETA_CLASS.NODE1_TYPE, node2_type=self.EDGE_INMETA_CLASS.NODE2_TYPE,
                         colindex_qscore=self.EDGE_INMETA_CLASS.QSCORE_COL,
                         mapping2_file=mapping_file2, map2_sourceindex=self.MAP2_META_CLASS.SOURCE_COL,
                         map2_targetindex=self.MAP2_META_CLASS.TARGET_COL,
                         altid_mapping2_file=altid_mapping_file2,
                         altid_map2_sourceindex=self.MAP2_ALT_ID_META_CLASS.SOURCE_COL,
                         altid_map2_targetindex=self.MAP2_ALT_ID_META_CLASS.TARGET_COL
                         )



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapUniEnsNcbi(InfileMetadata):
    CSV_NAME = "DB_Uniprot_mapping_gene_ensembl_ncbi.csv"
    USE_COLS = ['Ensembl', 'GeneID']
    SOURCE_COL = 0
    TARGET_COL = 1
    MAPPING_SEP = ";"
    INFILE_TYPE = InfileType.IN_MAP_UNI_ENS_NCBI

    def __init__(self):
        super().__init__(csv_name=InMetaMapUniEnsNcbi.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapUniEnsNcbi.INFILE_TYPE)



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapDrugCentralPubchem import InMetaMapDrugCentralPubchem
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class MapDrugCentralPubchemProcessor(FileProcessor):
    IN_META_CLASS = InMetaMapDrugCentralPubchem

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_MAP_DRUGCENTRAL_PUBCHEM,
                         infileType=InfileType.IN_MAP_DRUGCENTRAL_PUBCHEM, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_preprocessing(self, data):
        data = data[data.id_type == 'PUBCHEM_CID']
        return data



import csv
import json
import os

import openbiolink.graphProperties as graphProp
from openbiolink import globalConfig as globConst
from openbiolink.graph_creation import graphCreationConfig as gcConst


class GraphWriter():

    def __init__(self):
        self.graph_dir_path = os.path.join(globConst.WORKING_DIR, gcConst.GRAPH_FILES_FOLDER_NAME)
        os.makedirs(self.graph_dir_path, exist_ok=True)

    def output_graph_props(self):
        graph_prop_list = [item for item in dir(graphProp) if not item.startswith("__")]
        graph_prop_dict = {var: getattr(graphProp, var) for var in graph_prop_list}
        with open(os.path.join(self.graph_dir_path, 'graph_props.json'), 'w') as json_file:
            for k, v in graph_prop_dict.items():
                if not type(v) == str:
                    graph_prop_dict[k] = str(v)
            json.dump(graph_prop_dict, json_file, indent=4)

    @staticmethod
    def output_graph(nodes_dic: dict = None, edges_dic: dict = None, one_file_sep=None, multi_file_sep=None,
                     prefix=None, print_qscore=True, node_edge_list=True):
        if not prefix:
            prefix = ''

        if one_file_sep is None and multi_file_sep is None:
            one_file_sep = ','
        # one file
        if one_file_sep is not None:
            GraphWriter().output_graph_in_single_file(prefix=prefix, file_sep=one_file_sep, nodes_dic=nodes_dic,
                                                      edges_dic=edges_dic, qscore=print_qscore)
        # separate files
        if multi_file_sep is not None:
            GraphWriter().output_graph_in_multi_files(prefix, multi_file_sep, nodes_dic, edges_dic, qscore=print_qscore)
        # lists of all nodes and metaedges
        if node_edge_list:
            GraphWriter().write_node_and_edge_list(prefix, nodes_dic.keys(), edges_dic.keys())

        # niceToHave (8) adjacency matrix
        # key, value = nodes_dic
        # d = {x: i for i, x in enumerate(value)}
        # niceToHave (8) outputformat for graph DB

    def output_graph_in_single_file(self, prefix, file_sep, nodes_dic, edges_dic, qscore):
        if nodes_dic is not None:
            with open(os.path.join(self.graph_dir_path, prefix + gcConst.NODES_FILE_PREFIX + '.csv'), 'w') as out_file:
                writer = csv.writer(out_file, delimiter=file_sep, lineterminator='\n')
                for key, value in nodes_dic.items():
                    for node in value:
                        writer.writerow(list(node))
        if edges_dic is not None:
            with open(os.path.join(self.graph_dir_path, prefix + gcConst.EDGES_FILE_PREFIX + '.csv'), 'w') as out_file:
                writer = csv.writer(out_file, delimiter=file_sep, lineterminator='\n')
                for key, value in edges_dic.items():
                    for edge in value:
                        if qscore:
                            writer.writerow(list(edge))
                        else:
                            writer.writerow(edge.to_sub_rel_obj_list())

    def output_graph_in_multi_files(self, prefix, file_sep, nodes_dic, edges_dic, qscore):
        # write nodes
        for key, value in nodes_dic.items():
            with open(os.path.join(self.graph_dir_path, prefix + gcConst.NODES_FILE_PREFIX + '_' + key + '.csv'),
                      'w') as out_file:
                writer = csv.writer(out_file, delimiter=file_sep, lineterminator='\n')
                for node in value:
                    writer.writerow(list(node))
        # write edges
        for key, value in edges_dic.items():
            with open(os.path.join(self.graph_dir_path, prefix + gcConst.EDGES_FILE_PREFIX + '_' + key + '.csv'),
                      'w') as out_file:
                writer = csv.writer(out_file, delimiter=file_sep, lineterminator='\n')
                for edge in value:
                    if qscore:
                        writer.writerow(list(edge))
                    else:
                        writer.writerow(edge.to_sub_rel_obj_list())

    def write_node_and_edge_list(self, prefix, nodes_list, edges_list):
        with open(os.path.join(self.graph_dir_path, prefix + 'nodes_list.csv'), 'w') as out_file:
            out_file.writelines(list('\n'.join(nodes_list)))

        with open(os.path.join(self.graph_dir_path, prefix + 'edges_list.csv'), 'w') as out_file:
            out_file.writelines(list('\n'.join(edges_list)))



import os
import tkinter as tk
from tkinter import ttk, filedialog, messagebox

import openbiolink.evaluation.evalConfig as evalConst
import openbiolink.train_test_set_creation.ttsConfig as ttsConst
from openbiolink.evaluation.metricTypes import RankMetricType, ThresholdMetricType
from openbiolink.evaluation.models.modelTypes import ModelTypes
from openbiolink.gui import gui as gui


class EvalFrame(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller
        self.info_text = ""

        # title
        titles_panel = tk.Frame(self)
        self.info = tk.Button(titles_panel, text=" help ", command=lambda: gui.show_info_box(self.info_text))
        self.title = tk.Label(titles_panel, text="(3) Testing and Evaluation", font=controller.title_font)

        self.actions_el = self._create_action_el(self)

        threshold_metrics = [ThresholdMetricType[item] for item in dir(ThresholdMetricType) if
                             not item.startswith("__")]

        options_panel = tk.Frame(self)
        paths_box = tk.LabelFrame(options_panel, text='file paths')
        self.config_path_el = self._create_config_path_el(paths_box)
        self.train_file_el = self._create_train_path_el(paths_box)
        self.test_file_el = self._create_test_path_el(paths_box)
        self.trained_model_file_el = self._create_trained_model_path_el(paths_box)
        self.node_or_corrupted_file_el = self._create_nodes_or_corrupted_path_el(paths_box)

        self.metrics_box = tk.LabelFrame(options_panel, text='choose metrics')
        self.metrics_frame = tk.Frame(self.metrics_box)
        self.rank_metrics_dict = {}
        self.threshold_metrics_dict = {}

        self.hits_at_k_el = self._create_hits_at_k_el(self.metrics_frame)
        self.mrr_el = self._create_mrr_el(self.metrics_frame)
        self.hits_at_k_el.pack(side='top', fill='x', pady=(10, 5))
        self.mrr_el.pack(side='top', fill='x', pady=5)
        ttk.Separator(self.metrics_frame, orient='horizontal').pack(side='top', fill='x', pady=10, padx=10, anchor='s')
        for metric in threshold_metrics:
            self.threshold_metrics_dict[metric] = tk.BooleanVar(value=True)
            cb = tk.Checkbutton(self.metrics_frame, text=metric.value, variable=self.threshold_metrics_dict[metric])
            cb.pack(anchor='w', padx=5, pady=5)

        self.buttons_panel = tk.Frame(self)
        self.next_button = tk.Button(self.buttons_panel, text="Next", command=lambda: self.next_page(), height=1,
                                     width=15)
        self.prev_button = tk.Button(self.buttons_panel, text="Back",
                                     command=lambda: self.controller.show_previous_frame(), height=1, width=15)

        # packing
        titles_panel.pack(side="top", fill="x", pady=10)
        self.title.pack(side='left', pady=10, padx=15)
        self.info.pack(side="right", fill="x", pady=5, padx=15)
        self.actions_el.pack(side='top', fill='both', padx=15, pady=5, expand=True)

        options_panel.pack(side='top', fill='both', padx=15, pady=5, expand=True)
        self.metrics_box.pack(side='left', fill='both', expand=True, padx=(0, 5))
        self.metrics_frame.pack(side='top', fill='both', expand=True)
        paths_box.pack(side='left', fill='both', expand=True, padx=(5, 0))
        self.config_path_el.pack(side='top', fill='x')
        self.pack_file_paths()
        self.node_or_corrupted_file_el.pack(side='top', fill='x')

        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 1), padx=10, anchor='s')
        self.buttons_panel.pack(side='bottom', padx=15, fill='x')
        self.prev_button.pack(side='left', anchor='w', pady=(5, 10))
        self.next_button.pack(side='right', anchor='e', pady=(5, 10))

    def pack_file_paths(self):
        self.unpack_file_paths()
        if self.train.get():
            self.train_file_el.pack(side='top', fill='x')
        elif self.evaluate.get():
            self.trained_model_file_el.pack(side='top', fill='x')
        if self.evaluate.get():
            self.test_file_el.pack(side='top', fill='x')
            self.metrics_frame.pack(side='top', fill='both', expand=True)

        else:
            self.metrics_frame.pack_forget()
        if (self.hits.get() or self.mrr.get()) and self.evaluate.get():
            self.node_or_corrupted_file_el.pack(side='top', fill='x')

    def toggl_ranked_metrics(self):
        if not self.hits.get():
            self.rank_metrics_dict[RankMetricType.HITS_AT_K].set(False)
            self.rank_metrics_dict[RankMetricType.HITS_AT_K_UNFILTERED].set(False)
        if not self.mrr.get():
            self.rank_metrics_dict[RankMetricType.MRR].set(False)
            self.rank_metrics_dict[RankMetricType.MRR_UNFILTERED].set(False)

        self.pack_file_paths()

        if (self.hits.get() or self.mrr.get()) and self.evaluate.get():
            self.node_or_corrupted_file_el.pack(side='top', fill='x')

    def unpack_file_paths(self):
        self.train_file_el.pack_forget()
        self.trained_model_file_el.pack_forget()
        self.test_file_el.pack_forget()
        self.node_or_corrupted_file_el.pack_forget()

    def _create_action_el(self, parent):
        el = tk.LabelFrame(parent, text="general info")
        self.train = tk.BooleanVar(value=True)
        train_box = tk.Checkbutton(el, text='perform training', variable=self.train, command=self.pack_file_paths)
        self.evaluate = tk.BooleanVar(value=True)
        eval_box = tk.Checkbutton(el, text='perform evaluation', variable=self.evaluate, command=self.pack_file_paths)
        models = [ModelTypes[item].name for item in dir(ModelTypes) if
                  not item.startswith("__")]
        self.select_choices_models = models
        self.select_model = tk.StringVar()
        select_menu = tk.OptionMenu(el, self.select_model, *self.select_choices_models)
        select_menu.configure(width=20)
        train_box.pack(side='left', padx=5, anchor='w')
        eval_box.pack(side='left', padx=5, anchor='w')
        # ttk.Separator(el, orient='vertical').pack(side='left', fill='y', padx=10, anchor='w')
        select_menu.pack(side='right', padx=(5, 15), anchor='w')

        tk.Label(el, text='model:').pack(side='right')
        return el

    def _create_hits_at_k_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.hits = tk.BooleanVar(value=True)
        self.ks = tk.StringVar(value=str(evalConst.DEFAULT_HITS_AT_K))
        self.rank_metrics_dict[RankMetricType.HITS_AT_K] = tk.BooleanVar(value=True)
        self.rank_metrics_dict[RankMetricType.HITS_AT_K_UNFILTERED] = tk.BooleanVar(value=True)
        hits_cb = tk.Checkbutton(panel, text='hits@K', variable=self.hits, command=self.toggl_ranked_metrics)
        ks_entry = tk.Entry(panel, textvariable=self.ks)
        hits_filtered_cb = tk.Checkbutton(panel, text='filtered',
                                          variable=self.rank_metrics_dict[RankMetricType.HITS_AT_K])
        hits_unfiltered_cb = tk.Checkbutton(panel, text='unfiltered',
                                            variable=self.rank_metrics_dict[RankMetricType.HITS_AT_K_UNFILTERED])
        panel.pack(side='top', fill='x', padx=5)
        hits_cb.pack(side='left')
        hits_filtered_cb.pack(side='right')
        hits_unfiltered_cb.pack(side='right')
        ks_entry.pack(side='right')
        tk.Label(panel, text='K\'s:').pack(side='right')
        return el

    def _create_mrr_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.mrr = tk.BooleanVar(value=True)
        self.rank_metrics_dict[RankMetricType.MRR] = tk.BooleanVar(value=True)
        self.rank_metrics_dict[RankMetricType.MRR_UNFILTERED] = tk.BooleanVar(value=True)
        mrr_cb = tk.Checkbutton(panel, text='MRR', variable=self.mrr,
                                command=self.toggl_ranked_metrics)
        mrr_filtered_cb = tk.Checkbutton(panel, text='filtered',
                                         variable=self.rank_metrics_dict[RankMetricType.MRR])
        mrr_unfiltered_cb = tk.Checkbutton(panel, text='unfiltered',
                                           variable=self.rank_metrics_dict[RankMetricType.MRR_UNFILTERED])
        panel.pack(side='top', fill='x', padx=5)
        mrr_cb.pack(side='left')
        mrr_filtered_cb.pack(side='right')
        mrr_unfiltered_cb.pack(side='right')
        return el

    def _create_config_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.config_path = tk.StringVar()
        button = tk.Button(panel, text="select path ...", command=lambda: self.browse_config_file())
        label = tk.Entry(el, textvariable=self.config_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="model config path:").pack(side='left', anchor='w', padx=5, pady=5)
        button.pack(side='right', anchor='w', padx=5, pady=5)
        label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_train_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.train_path = tk.StringVar()
        train_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_train_file())
        train_label = tk.Entry(el, textvariable=self.train_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="training set path:").pack(side='left', anchor='w', padx=5, pady=5)
        train_button.pack(side='right', anchor='w', padx=5, pady=5)
        train_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_test_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.test_path = tk.StringVar()
        test_button = tk.Button(panel, text="select path ...", command=lambda: self.browse_test_file())
        test_label = tk.Entry(el, textvariable=self.test_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="test set path:").pack(side='left', anchor='w', padx=5, pady=5)
        test_button.pack(side='right', anchor='w', padx=5, pady=5)
        test_label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_trained_model_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.trained_model_path = tk.StringVar()
        button = tk.Button(panel, text="select path ...", command=lambda: self.browse_trained_model())
        label = tk.Entry(el, textvariable=self.trained_model_path)
        panel.pack(side='top', fill='x')
        tk.Label(panel, text="trained model path:").pack(side='left', anchor='w', padx=5, pady=5)
        button.pack(side='right', anchor='w', padx=5, pady=5)
        label.pack(fill='x', padx=5, pady=5)
        return el

    def _create_nodes_or_corrupted_path_el(self, parent):
        el = tk.Frame(parent)
        panel = tk.Frame(el)
        self.nodes_or_corr_path = tk.StringVar()
        button = tk.Button(panel, text="select path ...", command=lambda: self.browse_nodes_or_corr_file())
        self.select_choices_node_corrupted = ['nodes path', 'corrupted triples path']
        self.select_node_corrupted = tk.StringVar(value=self.select_choices_node_corrupted[0])
        select_menu = tk.OptionMenu(panel, self.select_node_corrupted, *self.select_choices_node_corrupted)

        select_menu.configure(width=20)
        label = tk.Entry(el, textvariable=self.nodes_or_corr_path)
        panel.pack(side='top', fill='x')
        select_menu.pack(side='left', anchor='w', padx=5, pady=5)
        button.pack(side='right', anchor='w', padx=5, pady=5)
        label.pack(fill='x', padx=5, pady=5)
        return el

    def browse_test_file(self):
        self.test_path.set(filedialog.askopenfilename())

    def browse_train_file(self):
        self.train_path.set(filedialog.askopenfilename())

    def browse_trained_model(self):
        self.trained_model_path.set(filedialog.askopenfilename())

    def browse_nodes_or_corr_file(self):
        self.nodes_or_corr_path.set(filedialog.askopenfilename())

    def browse_config_file(self):
        self.config_path.set(filedialog.askopenfilename())

    def update(self):
        # fixme check if graph creation action is performed
        tts_files_folder = os.path.join(self.controller.ARGS_LIST_GLOBAL[1], ttsConst.TTS_FOLDER_NAME)
        test_path = os.path.join(tts_files_folder, ttsConst.TEST_FILE_NAME)
        if os.path.exists(test_path) or 'SplitFrame' in self.controller.selected_frames:
            self.test_path.set(test_path)

        train_path = os.path.join(tts_files_folder, ttsConst.TRAIN_FILE_NAME)
        if os.path.exists(train_path) or 'SplitFrame' in self.controller.selected_frames:
            self.train_path.set(train_path)

        nodes_path = os.path.join(tts_files_folder, ttsConst.TRAIN_VAL_NODES_FILE_NAME)
        if os.path.exists(nodes_path) or 'GraphCreationFrame' in self.controller.selected_frames:
            self.nodes_or_corr_path.set(nodes_path)

    def next_page(self):
        if not self.train.get() and not self.evaluate.get():
            messagebox.showerror('ERROR', 'at least one action (training, evaluation) must be chosen')
            return
        if not self.select_model.get():
            messagebox.showerror('ERROR', 'Please select a model')
            return
        if self.train.get():
            if not self.train_path.get():
                messagebox.showerror('ERROR', 'For training, please select a training set path')
                return
        if self.evaluate.get():
            if not self.train.get() and not self.trained_model_path.get():
                messagebox.showerror('ERROR',
                                     'Please either provide a path to your trained model or choose \'perform training')
                return
            if not self.test_path.get():
                messagebox.showerror('ERROR', 'Please provide a trainings path')
                return
            if (self.hits.get() or self.mrr.get()) and not self.nodes_or_corr_path.get():
                messagebox.showerror('ERROR',
                                     'Ranked metrics (his@K, MRR) require either a nodes file or a file of corrupted triples')
                return
            if all([not self.mrr.get(), not self.hits.get()]) \
                    and all([not x.get() for _, x in self.threshold_metrics_dict.items()]):
                messagebox.showerror('ERROR',
                                     '\'Perform evaluation\' is chosen, but no evaluation metrics are selected')
                return
        self.controller.ARGS_LIST_EVAL = []
        self.controller.ARGS_LIST_EVAL.append('-e')
        self.controller.ARGS_LIST_EVAL.extend(['--model_cls', str(ModelTypes[self.select_model.get()].name)])
        if self.config_path.get():
            self.controller.ARGS_LIST_EVAL.extend(['--config', self.config_path.get()])
        if not self.train.get():
            self.controller.ARGS_LIST_EVAL.append('--no_train')
            self.controller.ARGS_LIST_EVAL.extend(['--trained_model', self.trained_model_path.get()])
        else:
            self.controller.ARGS_LIST_EVAL.extend(['--train', self.train_path.get()])
        if not self.evaluate.get():
            self.controller.ARGS_LIST_EVAL.append('--no_eval')
        else:
            self.controller.ARGS_LIST_EVAL.extend(['--test', self.test_path.get()])
            ranked_metrics = [x.name for x, y in self.rank_metrics_dict.items() if y.get()]
            threshold_metrics = [x.name for x, y in self.threshold_metrics_dict.items() if y.get()]
            self.controller.ARGS_LIST_EVAL.append('--metrics')
            self.controller.ARGS_LIST_EVAL.extend(ranked_metrics)
            self.controller.ARGS_LIST_EVAL.extend(threshold_metrics)
            if RankMetricType.HITS_AT_K in self.rank_metrics_dict.keys() or RankMetricType.HITS_AT_K_UNFILTERED in self.rank_metrics_dict.keys():
                self.controller.ARGS_LIST_EVAL.append('--ks')
                import ast
                self.controller.ARGS_LIST_EVAL.extend([str(x) for x in ast.literal_eval(self.ks.get())])
            # if metrics
            if self.select_node_corrupted.get() == 'nodes path':
                self.controller.ARGS_LIST_EVAL.extend(['--eval_nodes', self.nodes_or_corr_path.get()])
            elif self.select_node_corrupted.get() == 'corrupted triples path':
                self.controller.ARGS_LIST_EVAL.extend(['--corrupted', self.nodes_or_corr_path.get()])

        self.controller.show_next_frame()



from enum import Enum


class DbType(Enum):
    DB_ONTO_GO = 0
    DB_ONTO_DO = 1
    DB_ONTO_HPO = 2
    DB_ONTO_UBERON = 3

    DB_EDGE_CDT_PATH = 100
    DB_EDGE_DISGENET = 101
    DB_EDGE_GO = 102
    DB_EDGE_HPA = 103
    DB_EDGE_HPO_DIS = 104
    DB_EDGE_HPO_GENE = 105
    DB_EDGE_SIDER_IND = 106
    DB_EDGE_SIDER_SE = 107
    DB_EDGE_STITCH = 108
    DB_EDGE_STRING = 109
    DB_EDGE_DRUGCENTRAL = 110
    DB_EDGE_BGEE = 111
    DB_EDGE_BGEE_DIFF = 112
    DB_EDGE_STITCH_ACTION = 113
    DB_EDGE_STRING_ACTION = 114
    DB_EDGE_TN_HPO_DIS = 115

    DB_MAP_DISGENET = 10
    DB_MAP_STRING = 11
    DB_MAP_UNIPROT = 12



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeDrugCentral(DbMetadataEdge):
    NAME = 'Edge/Mapping - DrugCentral - (Contra)Indications, Mapping:(dc --> pubchem)'
    URL = "http://unmtid-shinyapps.net/download/drugcentral.dump.08262018.sql.gz"
    OFILE_NAME = "sql_dump.sql.gz"

    TABLE_NAME_IND = "public.omop_relationship"
    COLS_IND = [
        "id",
        "struct_id",
        "concept_id",
        "relationship_name",
        "concept_name",
        "umls_cui",
        "snomed_full_name",
        "cui_semantic_type",
        "snomed_conceptid"
    ]

    TABLE_NAME_MAP_PUBCHEM = "public.identifier"
    COLS_MAP_PUBCHEM = [
        "id",
        "identifier",
        "id_type",
        "struct_id",
        "parent_match"
    ]

    DB_TYPE = DbType.DB_EDGE_DRUGCENTRAL

    def __init__(self):
        super().__init__(url=DbMetaEdgeDrugCentral.URL,
                         ofile_name=DbMetaEdgeDrugCentral.OFILE_NAME,
                         dbType=DbMetaEdgeDrugCentral.DB_TYPE)



from openbiolink.graph_creation.metadata_db_file.mapping.dbMetaMapDisGeNet import DbMetaMapDisGeNet
from openbiolink.graph_creation.metadata_db_file.mapping.dbMetaMapString import DbMetaMapString
from openbiolink.graph_creation.metadata_db_file.mapping.dbMetaMapUniprot import DbMetaMapUniprot
from openbiolink.graph_creation.metadata_db_file.mapping.dbMetadataMapping import DbMetadataMapping



import os
import sys
import tkinter as tk
from tkinter import font  as tkfont, messagebox
from tkinter.ttk import Style

from openbiolink import openBioLink
from openbiolink.gui.confirmFrame import ConfirmFrame
from openbiolink.gui.console import ConsoleFrame
from openbiolink.gui.evaluationFrame import EvalFrame
from openbiolink.gui.graphCreationFrame import GraphCreationFrame
from openbiolink.gui.splitFrame import SplitFrame
from openbiolink.gui.startPage import StartPage

app = None


class BimegGui(tk.Tk):
    ARGS_LIST_GLOBAL = []
    ARGS_LIST_GRAPH_CREATION = []
    ARGS_LIST_TRAIN_TEST_SPLIT = []
    ARGS_LIST_EVAL = []

    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)

        # Define Fonts
        self.title_font = tkfont.Font(family='Helvetica', size=18, weight="bold", slant="italic")
        self.info_font = tkfont.Font(family='Helvetica', size=7, slant="italic")
        # Define base container
        self.container = tk.Frame(self)
        # self.wm_geometry('600x470')
        self.container.pack(side="top", fill="both", expand=True)
        self.container.grid_rowconfigure(0, weight=1)
        self.container.grid_columnconfigure(0, weight=1)

        # Initialize all frames
        self.frames = {}
        for F in (StartPage, GraphCreationFrame, SplitFrame, EvalFrame, ConfirmFrame, ConsoleFrame):
            page_name = F.__name__
            frame = F(parent=self.container, controller=self)
            self.frames[page_name] = frame

            frame.grid(row=0, column=0, sticky="nsew")

        self.args = []

        self.selected_frames = ['ConfirmFrame']
        self.next_frame_index = 0

        self.show_frame("StartPage")

    def get_args(self):
        arg_list = []
        arg_list.extend(self.ARGS_LIST_GLOBAL)
        arg_list.extend(self.ARGS_LIST_GRAPH_CREATION)
        arg_list.extend(self.ARGS_LIST_TRAIN_TEST_SPLIT)
        arg_list.extend(self.ARGS_LIST_EVAL)
        return arg_list

    def set_selected_frames(self, selected_frames):
        self.selected_frames = selected_frames + self.selected_frames

    def show_next_frame(self):
        if self.next_frame_index == len(self.selected_frames):
            self.start()
        else:
            self.show_frame(self.selected_frames[self.next_frame_index])
            self.next_frame_index += 1

    def show_previous_frame(self):
        if self.next_frame_index == 1:
            self.next_frame_index = 0
            self.show_frame("StartPage")
            self.selected_frames = ['ConfirmFrame']

        else:
            self.next_frame_index -= 1
            self.show_frame(self.selected_frames[self.next_frame_index - 1])

    def show_frame(self, page_name):
        """ Show a frame for the given page name """
        frame = self.frames[page_name]
        frame.controller = self
        frame.update()
        frame.tkraise()

    def start(self):
        """ start script and close gui"""
        if messagebox.askokcancel("Start", "Do you want to start now?"):
            self.show_frame("ConsoleFrame")
            arg_list = self.get_args()
            openBioLink.main(args_list=arg_list)
            # fixme start detached
            # app.destroy()


#################### ASK FOR EXIT ############################

class AskForExitPopup:
    def __init__(self, message):
        win = self.win = tk.Toplevel()

        l = tk.Label(self.win, text=message)
        l.grid(row=0, column=0)

        b1 = tk.Button(win, text="Cancel", command=self.cancel)
        b2 = tk.Button(win, text="Continue", command=self.go_on)
        b1.grid(row=1, column=0)
        b2.grid(row=1, column=1)
        win.wait_window()

    def cancel(self):
        self.win.destroy()
        return True

    def go_on(self):
        self.win.destroy()
        return False


def askForExit(message):
    exit = AskForExitPopup(message)
    if exit:
        on_closing()


#################### SKIP EXISTING FILES ############################

class SkipExistingFilesPopup:
    def __init__(self, file_path):
        self.skip = None
        self.for_all = False
        self.win = tk.Toplevel()
        message = 'The file %s already exists' % (file_path)
        l = tk.Label(self.win, text=message)

        button_panel = tk.Frame(self.win)

        go_on_button = tk.Button(button_panel, text="continue anyways", command=self.go_on)
        go_on_all_button = tk.Button(button_panel, text="continue anyways for all files", command=self.go_on_for_all)
        skip_button = tk.Button(button_panel, text="skip this file", command=self.skip_this)
        skip_all_button = tk.Button(button_panel, text="skip all existing files", command=self.skip_all)
        exit_button = tk.Button(button_panel, text="exit", command=self.exit)
        l.pack(side='top')
        button_panel.pack(side='top')
        go_on_button.pack(side='left')
        go_on_all_button.pack(side='left')
        skip_button.pack(side='left')
        skip_all_button.pack(side='left')
        exit_button.pack(side='left')
        self.win.wait_window()

    def exit(self):
        self.win.destroy()
        self.skip = None
        self.for_all = None

    def go_on(self):
        self.win.destroy()
        self.skip = None
        self.for_all = False

    def go_on_for_all(self):
        self.win.destroy()
        self.skip = False
        self.for_all = True

    def skip_this(self):
        self.win.destroy()
        self.skip = True
        self.for_all = False

    def skip_all(self):
        self.win.destroy()
        self.skip = True
        self.for_all = True


def skipExistingFiles(file_path):
    skip = None
    for_all = False
    if os.path.isfile(file_path):
        popup_response = SkipExistingFilesPopup(file_path)
        skip = popup_response.skip
        for_all = popup_response.for_all
        if skip is None and for_all is None:
            on_closing()
    return skip, for_all


#################### MAIN GUI ############################

def on_closing():
    if messagebox.askokcancel("Quit", "Do you really want to quit?"):
        app.destroy()
        sys.exit()


def show_info_box(msg):
    if not bool(msg):
        msg = 'No info available'
    messagebox.showinfo('Info', msg)


def start_gui():
    global app
    app = BimegGui()
    app.style = Style()

    app.protocol("WM_DELETE_WINDOW", on_closing)
    app.mainloop()



import gzip
import zipfile
from abc import ABC, abstractmethod


class FileReader(ABC):

    def __init__(self, in_path, readerType, dbType):
        self.in_path = in_path
        self.readerType = readerType
        self.dbType = dbType

    @staticmethod
    def open_file(in_path):
        path_parts = in_path.split('.')
        if (path_parts[-1] == "gz"):
            in_file = gzip.open(in_path, "rt", encoding="utf8")
        elif (path_parts[-1] == "zip"):
            zf = zipfile.ZipFile(in_path)
            in_file = zf.open(zf.namelist()[0])
        else:
            in_file = open(in_path, encoding="utf8")
        return in_file

    @abstractmethod
    def read_file(self):
        ...  # return pandas.Dataframe



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapUniUniNcbi(InfileMetadata):
    CSV_NAME = "DB_Uniprot_mapping_gene_uniprot_ncbi.csv"
    USE_COLS = ['UniProtKB-AC', 'GeneID']
    SOURCE_COL = 0
    TARGET_COL = 1
    MAPPING_SEP = ";"
    INFILE_TYPE = InfileType.IN_MAP_UNI_UNI_NCBI

    def __init__(self):
        super().__init__(csv_name=InMetaMapUniUniNcbi.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapUniUniNcbi.INFILE_TYPE)



import os
import sys


class Cli():

    @staticmethod
    def ask_for_exit(message):
        user_input = input(message + '\nDo you want to \n'
                                     ' [c] continue \n'
                                     ' [x] exit \n')
        if user_input == 'x':
            sys.exit()

    @staticmethod
    def skip_existing_files(file_path):
        skip = None
        for_all = False
        if os.path.isfile(file_path):
            user_input = input(
                '\nThe file %s already exists. \n'
                'Do you want to \n'
                ' [y] continue anyways\n'
                ' [c] continue anyways for all files\n'
                ' [n] skip this file\n'
                ' [s] skip all existing files\n'
                ' [x] exit \n' % (file_path))
            if user_input == 'x':
                sys.exit()
            elif user_input == 'c':
                skip = False
                for_all = True
            elif user_input == 'n':
                skip = True
            elif user_input == 's':
                skip = True
                for_all = True
        return skip, for_all



from enum import Enum


class EdgeType(Enum):
    IS_A = 0
    PART_OF = 1

    GENE_GENE = 10
    GENE_GO = 11
    GENE_DIS = 12
    GENE_DRUG = 13
    GENE_PHENOTYPE = 14
    GENE_PATHWAY = 15
    GENE_EXPRESSED_ANATOMY = 16
    DIS_DRUG = 17
    DIS_PHENOTYPE = 18
    DRUG_PHENOTYPE = 19
    GENE_OVEREXPRESSED_ANATOMY = 20
    GENE_UNDEREXPRESSED_ANATOMY = 201

    DRUG_ACTIVATION_GENE = 21
    DRUG_BINDING_GENE = 22
    DRUG_CATALYSIS_GENE = 23
    DRUG_EXPRESSION_GENE = 24
    DRUG_INHIBITION_GENE = 25
    DRUG_PREDBIND_GENE = 26
    DRUG_REACTION_GENE = 27
    DRUG_BINDACT_GENE = 28
    DRUG_BINDINH_GENE = 29

    GENE_ACTIVATION_GENE = 31
    GENE_BINDING_GENE = 32
    GENE_CATALYSIS_GENE = 33
    GENE_EXPRESSION_GENE = 34
    GENE_INHIBITION_GENE = 35
    GENE_PTMOD_GENE = 36
    GENE_REACTION_GENE = 37
    GENE_BINDACT_GENE = 38
    GENE_BINDINH_GENE = 39

    def __str__(self):
        return (self.name)

    def get_parent(self):
        if self in [EdgeType.DRUG_ACTIVATION_GENE,
                    EdgeType.DRUG_BINDING_GENE,
                    EdgeType.DRUG_CATALYSIS_GENE,
                    EdgeType.DRUG_EXPRESSION_GENE,
                    EdgeType.DRUG_INHIBITION_GENE,
                    EdgeType.DRUG_PREDBIND_GENE,
                    EdgeType.DRUG_REACTION_GENE,
                    EdgeType.DRUG_BINDACT_GENE,
                    EdgeType.DRUG_BINDINH_GENE]:
            return EdgeType.GENE_DRUG
        elif self in [EdgeType.GENE_ACTIVATION_GENE,
                      EdgeType.GENE_BINDING_GENE,
                      EdgeType.GENE_CATALYSIS_GENE,
                      EdgeType.GENE_EXPRESSION_GENE,
                      EdgeType.GENE_INHIBITION_GENE,
                      EdgeType.GENE_PTMOD_GENE,
                      EdgeType.GENE_REACTION_GENE,
                      EdgeType.GENE_BINDACT_GENE,
                      EdgeType.GENE_BINDINH_GENE]:
            return EdgeType.GENE_GENE
        elif self in [EdgeType.GENE_OVEREXPRESSED_ANATOMY,
                      EdgeType.GENE_UNDEREXPRESSED_ANATOMY]:
            return EdgeType.GENE_EXPRESSED_ANATOMY
        else:
            return self



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapUniUniNcbi import InMetaMapUniUniNcbi
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class MapUniUniNcbiProcessor(FileProcessor):
    IN_META_CLASS = InMetaMapUniUniNcbi

    def __init__(self):
        super().__init__(self.IN_META_CLASS.USE_COLS, readerType=ReaderType.READER_MAP_UNIPROT,
                         infileType=InfileType.IN_MAP_UNI_UNI_NCBI, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)



from openbiolink import utils
from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile import InMetaEdgeStringBinding
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class EdgeStringBindingProcessor(FileProcessor):
    IN_META_CLASS = InMetaEdgeStringBinding

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_EDGE_STRING_ACTION,
                         infileType=InfileType.IN_EDGE_STRING_BINDING, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_preprocessing(self, data):
        # binding is an undirectional link --> take only the undirectional cases and only in one direction
        # data = data[data['action'].isnull()]
        data = data[data['mode'] == 'binding']
        data = data[data['is_directional'] == 'f']
        return data

    def individual_postprocessing(self, data):
        return utils.make_undir(data)



from openbiolink.graph_creation.metadata_infile.infileMetadata import InfileMetadata
from openbiolink.graph_creation.types.infileType import InfileType


class InMetaMapDrugCentralPubchem(InfileMetadata):
    CSV_NAME = "DB_DrugCentral_mapping_drug_pubchem.csv"

    USE_COLS = [
        "identifier",
        "struct_id",
    ]
    SOURCE_COL = 1
    TARGET_COL = 0
    MAPPING_SEP = None
    INFILE_TYPE = InfileType.IN_MAP_DRUGCENTRAL_PUBCHEM

    def __init__(self):
        super().__init__(csv_name=InMetaMapDrugCentralPubchem.CSV_NAME,
                         cols=self.USE_COLS,
                         infileType=InMetaMapDrugCentralPubchem.INFILE_TYPE)



import os
import tkinter as tk
from tkinter import messagebox, filedialog, ttk

from openbiolink.gui import gui as gui


class StartPage(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller
        self.info_text = "select "

        titles_panel = tk.Frame(self)
        info = tk.Button(titles_panel, text=" help ", command=lambda: gui.show_info_box(self.info_text))
        title = tk.Label(titles_panel, text="Select Actions", font=controller.title_font)

        self.working_dir_el = self._create_working_dir_el(self)

        actions_panel = tk.LabelFrame(self, text='Actions')
        self.g = tk.BooleanVar()
        g_box = tk.Checkbutton(actions_panel, text="(1) Generate Graph", variable=self.g)
        self.s = tk.BooleanVar()
        s_box = tk.Checkbutton(actions_panel, text="(2) Generate Train Test Split", variable=self.s)
        self.e = tk.BooleanVar()
        e_box = tk.Checkbutton(actions_panel, text="(3) Apply Testing and Evaluation", variable=self.e)

        buttons_panel = tk.Frame(self)
        next_button = tk.Button(buttons_panel, text="Next", command=lambda: self.next_page(), height=1, width=15)

        titles_panel.pack(side="top", fill="x", pady=10)
        title.pack(side='left', pady=10, padx=15)
        info.pack(side="right", fill="x", pady=10, padx=15)
        self.working_dir_el.pack(side='top', fill='both', padx=15, pady=10, expand=True)
        actions_panel.pack(side='top', fill='both', expand=True, anchor='w', padx=15, pady=10)
        g_box.pack(side='top', anchor='w', padx=20, pady=(20, 0))
        s_box.pack(side='top', anchor='w', padx=20, pady=(20, 0))
        e_box.pack(side='top', anchor='w', padx=20, pady=(20, 0))
        ttk.Separator(self, orient='horizontal').pack(side='top', fill='x', pady=(15, 0), padx=10, anchor='s')
        buttons_panel.pack(side='bottom', padx=15, fill='x')
        next_button.pack(side='right', anchor='e', pady=(5, 10))

    def _create_working_dir_el(self, parent):
        el = tk.LabelFrame(parent, text="Working Directory")
        self.path = tk.StringVar(value=os.getcwd())  # todo good idea?
        path_button = tk.Button(el, text="select path ...", command=lambda: self.browse_dir())
        path_label = tk.Label(el, textvariable=self.path)
        path_button.pack(side='left', anchor='w', padx=5, pady=5)
        path_label.pack(side='left', anchor='w', padx=5, pady=5)
        return el

    def browse_dir(self):
        self.path.set(filedialog.askdirectory())

    def next_page(self):
        self.controller.ARGS_LIST_GLOBAL = []
        if self.path.get() == '':
            messagebox.showerror('ERROR', 'Please select a path')
            return
        self.controller.ARGS_LIST_GLOBAL.extend(['-p', self.path.get()])

        selected_frames = []
        if self.g.get():
            selected_frames.append("GraphCreationFrame")
        else:
            self.ARGS_LIST_GRAPH_CREATION = []
        if self.s.get():
            selected_frames.append("SplitFrame")
        else:
            self.controller.ARGS_LIST_TRAIN_TEST_SPLIT = []
        if self.e.get():
            selected_frames.append("EvalFrame")
        else:
            self.controller.ARGS_LIST_EVAL = []
        if len(selected_frames) < 1:
            messagebox.showerror("ERROR", "At least one action must be chosen!")
            return
        self.controller.set_selected_frames(selected_frames)
        self.controller.show_next_frame()



from openbiolink.graph_creation.metadata_db_file.onto.dbMetadataOnto import DbMetadataOnto
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaOntoHpo(DbMetadataOnto):
    NAME = 'Onto - HPO - is_a, mapping(umls->hpo)'
    URL = "https://raw.githubusercontent.com/obophenotype/human-phenotype-ontology/master/hp.obo"
    OFILE_NAME = "HPO_ontology.obo"
    QUADRUPLES = [('id', ' ', 1, 'ID'),
                  ('alt_id', ' ', 1, 'ALT_ID'),
                  ('is_a', ' ', 1, 'IS_A'),
                  ('xref: UMLS:', ':', 2, 'UMLS')]  # mapping to UMLS
    DB_TYPE = DbType.DB_ONTO_HPO

    def __init__(self):
        super().__init__(url=DbMetaOntoHpo.URL,
                         ofile_name=DbMetaOntoHpo.OFILE_NAME,
                         dbType=DbMetaOntoHpo.DB_TYPE)



from openbiolink.graph_creation.file_processor.fileProcessor import FileProcessor
from openbiolink.graph_creation.metadata_infile import InMetaEdgeStringActivation
from openbiolink.graph_creation.types.infileType import InfileType
from openbiolink.graph_creation.types.readerType import ReaderType


class EdgeStringActivationProcessor(FileProcessor):
    IN_META_CLASS = InMetaEdgeStringActivation

    def __init__(self):
        self.use_cols = self.IN_META_CLASS.USE_COLS
        super().__init__(self.use_cols, readerType=ReaderType.READER_EDGE_STRING_ACTION,
                         infileType=InfileType.IN_EDGE_STRING_ACTIVATION, mapping_sep=self.IN_META_CLASS.MAPPING_SEP)

    def individual_preprocessing(self, data):
        # activation is a directional link --> take only the directional cases where a is acting
        data = data[data['mode'] == 'activation']
        data = data[data['is_directional'] == 't']
        data = data[data['a_is_acting'] == 't']
        return data



import os

from openbiolink.graph_creation import graphCreationConfig as glob
from openbiolink.graph_creation.metadata_edge.edgeRegularMetadata import EdgeRegularMetadata
from openbiolink.graph_creation.metadata_infile import InMetaEdgeBgeeUnderExpr, InMetaMapOntoUberonAltid
from openbiolink.graph_creation.metadata_infile.mapping.inMetaMapUniEnsNcbi import InMetaMapUniEnsNcbi
from openbiolink.graph_creation.types.qualityType import QualityType


class EdgeMetaGeneUnderAna(EdgeRegularMetadata):
    NAME = 'Edge - Gene_underExpression_Anatomy'

    LQ_CUTOFF_TEXT = None
    MQ_CUTOFF_TEXT = None
    HQ_CUTOFF_TEXT = ['low quality']

    EDGE_INMETA_CLASS = InMetaEdgeBgeeUnderExpr
    MAP1_META_CLASS = InMetaMapUniEnsNcbi
    MAP2_ALT_ID_META_CLASS = InMetaMapOntoUberonAltid

    def __init__(self, quality: QualityType = None):
        edges_file_path = os.path.join(glob.IN_FILE_PATH, self.EDGE_INMETA_CLASS.CSV_NAME)
        mapping_file1 = os.path.join(glob.IN_FILE_PATH, self.MAP1_META_CLASS.CSV_NAME)
        altid_mapping2_file = os.path.join(glob.IN_FILE_PATH, self.MAP2_ALT_ID_META_CLASS.CSV_NAME)
        super().__init__(is_directional=True,
                         edges_file_path=edges_file_path,
                         colindex1=self.EDGE_INMETA_CLASS.NODE1_COL, colindex2=self.EDGE_INMETA_CLASS.NODE2_COL,
                         edgeType=self.EDGE_INMETA_CLASS.EDGE_TYPE,
                         node1_type=self.EDGE_INMETA_CLASS.NODE1_TYPE, node2_type=self.EDGE_INMETA_CLASS.NODE2_TYPE,
                         colindex_qscore=self.EDGE_INMETA_CLASS.QSCORE_COL, quality=quality,
                         mapping1_file=mapping_file1, map1_sourceindex=self.MAP1_META_CLASS.SOURCE_COL,
                         map1_targetindex=self.MAP1_META_CLASS.TARGET_COL, altid_mapping2_file=altid_mapping2_file,
                         altid_map2_sourceindex=self.MAP2_ALT_ID_META_CLASS.SOURCE_COL,
                         altid_map2_targetindex=self.MAP2_ALT_ID_META_CLASS.TARGET_COL
                         )



from openbiolink.graph_creation.metadata_db_file.onto.dbMetaOntoDo import DbMetaOntoDo
from openbiolink.graph_creation.metadata_db_file.onto.dbMetaOntoGo import DbMetaOntoGo
from openbiolink.graph_creation.metadata_db_file.onto.dbMetaOntoHpo import DbMetaOntoHpo
from openbiolink.graph_creation.metadata_db_file.onto.dbMetaOntoUberon import DbMetaOntoUberon
from openbiolink.graph_creation.metadata_db_file.onto.dbMetadataOnto import DbMetadataOnto



from abc import ABC, abstractmethod

import numpy as np


class Model(ABC):

    @abstractmethod
    def __init__(self):
        ...

    @abstractmethod
    def train(self, pos_triples: np.array, neg_triples: np.array):
        ...

    @abstractmethod
    def get_ranked_and_sorted_predictions(self, examples):
        ...

    @abstractmethod
    def output_model(self, path):
        ...

    def _split_list_in_batches(self, input_list, batch_size):
        return [input_list[i:i + batch_size] for i in range(0, len(input_list), batch_size)]

# todo check if num embedding == #nodes



from openbiolink.graph_creation.metadata_db_file.edge.dbMetadataEdge import DbMetadataEdge
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaEdgeSiderInd(DbMetadataEdge):
    NAME = 'Edge - Sider - Indications'
    URL = "http://sideeffects.embl.de/media/download/meddra_all_indications.tsv.gz"
    OFILE_NAME = "SIDER_dis_drug.tsv.gz"
    COLS = ['stichID', 'umlsID', 'method', 'umlsName', 'medDRAumlsType',
            'medDRAumlsID', 'medDRAumlsName']
    FILTER_COLS = ['umlsID', 'stichID', 'method']
    HEADER = 0
    DB_TYPE = DbType.DB_EDGE_SIDER_IND

    def __init__(self):
        super().__init__(url=DbMetaEdgeSiderInd.URL,
                         ofile_name=DbMetaEdgeSiderInd.OFILE_NAME,
                         dbType=DbMetaEdgeSiderInd.DB_TYPE)



import argparse
import cProfile
import logging
import os
import sys

import openbiolink.graphProperties as graphProp
from openbiolink import globalConfig
from openbiolink import globalConfig as glob
from openbiolink.evaluation.evaluation import Evaluation
from openbiolink.evaluation.metricTypes import RankMetricType, ThresholdMetricType
from openbiolink.evaluation.models.modelTypes import ModelTypes
from openbiolink.graph_creation.graphCreation import Graph_Creation
from openbiolink.graph_creation.types.qualityType import QualityType
from openbiolink.train_test_set_creation.trainTestSplitCreation import TrainTestSetCreation


def create_graph(args):
    graphProp.DIRECTED = not args.undir
    graphProp.QUALITY = args.qual
    globalConfig.INTERACTIVE_MODE = not args.no_interact
    globalConfig.SKIP_EXISTING_FILES = args.skip

    use_db_metadata_classes = None
    use_edge_metadata_classes = None
    # todo via enums
    if args.dbs:
        db_module_names = ['.'.join(y) for y in [x.split('.')[0:-1] for x in args.dbs]]
        db_cls_names = [x.split('.')[-1] for x in args.dbs]
        use_db_metadata_classes = [getattr(sys.modules[module_name], cls_name) for module_name, cls_name in
                                   zip(db_module_names, db_cls_names)]
    if args.mes:
        db_module_names = ['.'.join(y) for y in [x.split('.')[0:-1] for x in args.mes]]
        db_cls_names = [x.split('.')[-1] for x in args.mes]
        use_edge_metadata_classes = [getattr(sys.modules[module_name], cls_name) for module_name, cls_name in
                                     zip(db_module_names, db_cls_names)]

    graph_creator = Graph_Creation(folder_path=glob.WORKING_DIR,
                                   use_db_metadata_classes=use_db_metadata_classes,
                                   use_edge_metadata_classes=use_edge_metadata_classes)

    logging.info('###### (1) GRAPH CREATION ######')
    if not args.no_dl:
        print("\n\n############### downloading files #################################")
        logging.info('## Start downloading files ##')
        graph_creator.download_db_files()

    if not args.no_in:
        print("\n\n############### creating graph input files #################################")
        logging.info('## Start creating input files ##')
        graph_creator.create_input_files()

    if not args.no_create:
        print("\n\n############### creating graph #################################")
        logging.info('## Start creating graph ##')
        if 's' in args.out_format[0]:
            single_sep = args.out_format[1][args.out_format[0].index('s')]
            if single_sep == 'n' or single_sep == 't':
                single_sep = single_sep.replace('n', '\n').replace('t', '\t')
        else:
            single_sep = None
        if 'm' in args.out_format[0]:
            multisep_sep = args.out_format[1][args.out_format[0].index('m')]
            if multisep_sep == 'n' or multisep_sep == 't':
                multisep_sep = multisep_sep.replace('n', '\n').replace('t', '\t')
        else:
            multisep_sep = None

        graph_creator.create_graph(one_file_sep=single_sep, multi_file_sep=multisep_sep,
                                   print_qscore=(not args.no_qscore))

        # with open(os.path.join(globalConfig.WORKING_DIR, globalConfig.GRAPH_PROP_FILE_NAME), 'w') as f:
        #    graph_prop_dict  = {x: y for x, y in graphProp.__dict__.items() if not x.startswith('__')}
        #    for k,v in  graph_prop_dict.items():
        #        if not type(v)==str:
        #            graph_prop_dict[k] = str(v)
        #    json.dump(graph_prop_dict, f, indent=4)


def create_train_test_splits(args):
    if args.tts_sep == 't':
        sep = '\t'
    elif args.tts_sep == 'n':
        sep = '\n'
    else:
        sep = args.tts_sep
    tts = TrainTestSetCreation(graph_path=args.edges,
                               tn_graph_path=args.tn_edges,
                               all_nodes_path=args.nodes,
                               sep=sep,
                               # meta_edge_triples=args.meta,
                               t_minus_one_graph_path=args.tmo_edges,
                               t_minus_one_tn_graph_path=args.tmo_tn_edges,
                               t_minus_one_nodes_path=args.tmo_nodes)
    if args.mode == 'time':
        print("\n\n############### creating time slice split #################################")
        logging.info('## Start creating time slice split ##')
        tts.time_slice_split()
    elif args.mode == 'rand':
        print("\n\n############### creating random slice split #################################")
        logging.info('## Start creating random slice split ##')
        tts.random_edge_split(val=args.val, test_frac=args.test_frac, crossval=args.crossval)
    # tts.random_edge_split(crossval=False)


def train_and_evaluate(args):
    model_cls = ModelTypes[args.model_cls].value
    if args.trained_model:  # fixme when model provided, config also has to be
        if args.config:
            config = args.config
            # model = model_cls(args.config)
        else:
            config = None
        model = model_cls()
        model.kge_model = model_cls.load_model(config)
        # model.kge_model.load_state_dict(torch.load(args.trained_model))
        # testme
    elif args.config:
        model = model_cls(args.config)
    else:
        model = model_cls()
    e = Evaluation(model=model, training_set_path=args.train, test_set_path=args.test, nodes_path=args.eval_nodes,
                   mappings_avail=bool(args.trained_model))
    # todo doku: mappings have to be in model folder with correct name, or change here
    if not args.no_train:
        print('starting training')
        e.train()
    if not args.no_eval:
        print('starting evaluation')

        metric_strings = args.metrics
        metrics = [x for x in list(RankMetricType.__members__.values()) if x.name in metric_strings] + \
                  [x for x in list(ThresholdMetricType.__members__.values()) if x.name in metric_strings]
        int_ks = [int(k) for k in args.ks]

        e.evaluate(metrics=metrics, ks=int_ks)


def check_args_validity(args, parser):
    # global checks
    if not (args.g or args.s or args.e):
        parser.error("at least one action is required [-g, -s, -e]")
    if args.skip and args.no_interact is None:
        parser.error("option --skip requires --no_interact")
    # graph creation checks
    if args.g:
        if args.no_in and not args.no_dl and not args.no_create:
            parser.error(
                "Graph Creation: downloading graph files and creating the graph without creating in_files is not possible")
    # train test split checks
    if args.s:
        if not args.edges or not args.tn_edges or not args.nodes:
            parser.error("Train Test Split: paths to the edge file (--edges), negative edge file (--tn_edges) "
                         "and nodes file (--nodes) must be provided with option -s")
        if args.crossval:
            n_folds = args.val
            if n_folds == 0 or n_folds == 1 or (n_folds > 1 and not float(n_folds).is_integer()):
                parser.error(
                    "fold entry must be either an int>1 (number of folds) or a float >0 and <1 (validation fraction)")
        if args.mode == 'time':
            if not (bool(args.tmo_edges) and bool(args.tmo_tn_edges) and (bool(args.tmo_nodes))):
                parser.error(
                    'for time slice mode, edge-, tn-edge- and node-file of the t-minus-one graph must be provided')
    # if args.crossval and (not args.tmo_edges or not args.tmo_tn_edges or not args.tmo_nodes):
    #            parser.error("Train Test Split: paths to the t-1 edge file (--tmo_edges),"
    #                         " t-1 negative edge file (--tmo_tn_edges) and t-1 nodes file (--tmo_nodes) "
    #                         "must be provided with option --crossval") #todo what?


def main(args_list=None):
    if (len(sys.argv) < 2) and not args_list:
        glob.GUI_MODE = True
        import openbiolink.gui.gui as gui
        gui.start_gui()
        return

    parser = argparse.ArgumentParser('OpenBioLink Toolbox')

    # Global config
    parser.add_argument('-p', type=str, default=os.getcwd(),
                        help='specify a working directory (default = current working dictionary')

    # Graph Creation
    parser.add_argument('-g', action='store_true', help='Generate Graph')
    parser.add_argument('--undir', action='store_true',
                        help='Output-Graph should be undirectional (default = directional)')
    parser.add_argument('--qual', type=str,
                        help='quality level od the output-graph, options = [hq, mq, lq], (default = None -> all entries are used)')
    parser.add_argument('--no_interact', action='store_true',
                        help='Disables interactive mode - existing files will be replaced (default = interactive)')
    parser.add_argument('--skip', action='store_true',
                        help='Existing files will be skipped - in combination with --no_interact (default = replace)')
    parser.add_argument('--no_dl', action='store_true',
                        help='No download is being performed (e.g. when local data is used)')
    parser.add_argument('--no_in', action='store_true',
                        help='No input_files are created (e.g. when local data is used)')
    parser.add_argument('--no_create', action='store_true',
                        help='No graph is created (e.g. when only in-files should be created)')
    parser.add_argument('--out_format', nargs=2, type=list, default='s t',
                        help='Format of graph output, takes 2 arguments: list of file formats [s= single file, m=multiple files] and list of separators (e.g. t=tab, n=newline, or any other character) (default= s t)')
    parser.add_argument('--no_qscore', action='store_true', help='The output files will contain no scores')
    parser.add_argument('--dbs', nargs='+',
                        help='custom source databases selection to be used, full class name, options --> see doc')
    parser.add_argument('--mes', nargs='+',
                        help='custom meta edges selection to be used, full class name, options --> see doc')

    # Train- Test Split Generation
    parser.add_argument('-s', action='store_true', help='Generate Train-,Validation-, Test-Split')
    parser.add_argument('--edges', type=str, help='Path to edges.csv file (required with action -s')
    parser.add_argument('--tn_edges', type=str, help='Path to true_negatives_edges.csv file (required with action -s)')
    parser.add_argument('--nodes', type=str, help='Path to nodes.csv file (required with action -s)')
    parser.add_argument('--tts_sep', type=str, default='t',
                        help='Separator of edge, tn-edge and nodes file (e.g. t=tab, n=newline, or any other character) (default=t)')
    parser.add_argument('--mode', type=str, default='rand',
                        help='Mode of train-test-set split, options=[rand, time], (default=rand)')
    parser.add_argument('--test_frac', type=float, default='0.2', help='Fraction of test set as float (default= 0.2)')
    parser.add_argument('--crossval', action='store_true', help='Multiple train-validation-sets are generated')
    parser.add_argument('--val', type=float, default='0.2',
                        help='Fraction of validation set as float (default= 0.2) or number of folds as int')
    # niceToHave (1)
    # parser.add_argument('--meta', type=str, help='Path to meta_edge triples (only required if meta-edges are not in OpenBioLink Benchmark Data)')
    parser.add_argument('--tmo_edges', type=str,
                        help='Path to edges.csv file of t-minus-one graph (required for --mode time')
    parser.add_argument('--tmo_tn_edges', type=str,
                        help='Path to true_negatives_edges.csv file of t-minus-one graph (required for --mode time')
    parser.add_argument('--tmo_nodes', type=str,
                        help='Path to nodes.csv file of t-minus-one graph (required for --mode time')

    # Training and Evaluation
    parser.add_argument('-e', action='store_true', help='Apply Training and Evaluation')
    parser.add_argument('--model_cls', type=str, help='class of the model to be trained/evaluated (required with -e)')
    parser.add_argument('--config', type=str, help='Path to the model\' config file')
    parser.add_argument('--no_train', action='store_true',
                        help='No training is being performed, trained model id provided via --model')
    parser.add_argument('--trained_model', type=str, help='Path to trained model (required with --no_train)')
    parser.add_argument('--no_eval', action='store_true', help='No evaluation is being performed, only training')
    parser.add_argument('--test', type=str, help='Path to test set file (required with -e)')
    parser.add_argument('--train', type=str, help='Path to trainings set file')  # (alternative: --cv_folder)')
    parser.add_argument('--corrupted', type=str,
                        help='path to the corrupted triples (required for ranked triples if no nodes file is provided')  # fixme no longer an option
    parser.add_argument('--eval_nodes', type=str,
                        help='path to the nodes file (required for ranked triples if no corrupted triples file is provided and nodes cannot be taken from graph creation')
    parser.add_argument('--metrics', nargs='+', help='evaluation metrics')
    parser.add_argument('--ks', nargs='+', help='k\'s for hits@k metric')
    # parser.add_argument('--cv_folder', type=str, help='Path to cross validation folder (alternative: --train)')

    # niceToHave (7) option to load info from config file

    if args_list:
        args = parser.parse_args(args_list)
    else:
        args = parser.parse_args()

    check_args_validity(args, parser)
    glob.WORKING_DIR = args.p

    if args.qual == 'hq':
        args.qual = QualityType.HQ
    elif args.qual == 'mq':
        args.qual = QualityType.MQ
    elif args.qual == 'lq':
        args.qual = QualityType.LQ

    if args.g:
        create_graph(args)
    if args.s:
        create_train_test_splits(args)
    if args.e:
        train_and_evaluate(args)

    logging.info('Finished!')


if __name__ == '__main__':
    logger = logging.getLogger()
    logger.setLevel('INFO')
    pr = cProfile.Profile()
    pr.enable()
    main()
    base_dir = os.path.dirname(os.path.realpath(__file__))
    test_folder = os.path.join(base_dir, 'test')

    pr.disable()
    pr.print_stats(sort="time")

    # todo crossval eval



from openbiolink.graph_creation.metadata_db_file.onto.dbMetadataOnto import DbMetadataOnto
from openbiolink.graph_creation.types.dbType import DbType


class DbMetaOntoGo(DbMetadataOnto):
    NAME = 'Onto - GO - is_a,part_of'
    URL = "http://purl.obolibrary.org/obo/go/go-basic.obo"
    OFILE_NAME = "GO_ontology.obo"
    QUADRUPLES = [('id', ' ', 1, 'ID'),
                  ('alt_id', ' ', 1, 'ALT_ID'),
                  ('is_a', ' ', 1, 'IS_A'),
                  ('relationship: part_of', ' ', 2, 'PART_OF')]
    DB_TYPE = DbType.DB_ONTO_GO

    def __init__(self):
        super().__init__(url=DbMetaOntoGo.URL,
                         ofile_name=DbMetaOntoGo.OFILE_NAME,
                         dbType=DbMetaOntoGo.DB_TYPE)



from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoDoIsA import InMetaOntoDoIsA
from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoGoIsA import InMetaOntoGoIsA
from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoGoPartOf import InMetaOntoGoPartOf
from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoHpoIsA import InMetaOntoHpoIsA
from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoUberonIsA import InMetaOntoUberonIsA
from openbiolink.graph_creation.metadata_infile.onto.inMetaOntoUberonPartOf import InMetaOntoUberonPartOf



from openbiolink.graph_creation.file_processor.onto.ontoDoIsAProcessor import OntoDoIsAProcessor
from openbiolink.graph_creation.file_processor.onto.ontoGoIsAProcessor import OntoGoIsAProcessor
from openbiolink.graph_creation.file_processor.onto.ontoGoPartOfProcessor import OntoGoPartOfProcessor
from openbiolink.graph_creation.file_processor.onto.ontoHpoIsAProcessor import OntoHpoIsAProcessor
from openbiolink.graph_creation.file_processor.onto.ontoUberonIsAProcessor import OntoUberonIsAProcessor
from openbiolink.graph_creation.file_processor.onto.ontoUberonPartOfProcessor import OntoUberonPartOfProcessor



from openbiolink.graph_creation.file_reader.fileReader import FileReader
from openbiolink.graph_creation.file_reader.parser.postgresDumpParser import PostgresDumpParser as dcp


class PostgresDumpReader(FileReader):

    def __init__(self, in_path, table_name, cols, readerType, dbType):
        super().__init__(in_path, readerType, dbType)
        self.table_name = table_name
        self.cols = cols

    def read_file(self):
        with FileReader.open_file(self.in_path) as file:
            df = dcp.table_to_df(file, self.table_name, self.cols)
            return df
